from os.path import basename
from os.path import exists
from uuid import uuid1

from lk_logger import lk

from .path_formatter import PathFormatter
from ...path_model import prj_model
from ...typehint import *


def indexing_paths(conf: TConf, path_fmt: Union[PathFormatter, Callable]):
    """
    Args:
        conf
        path_fmt:
            - PathFormatter: map path node from dir_i to dir_o.
                - the default dir_i is dirname of pyproj_file.
                - the default dir_o is dist dir from conf['build']['dist_dir'].
            - Callable: receives one argument as path_i, returns path_o. the
                typical pattern is: `lambda path_i: path_o`.
    """
    placeholders = {
        'app_name'       : conf['app_name'],
        'app_name_lower' : conf['app_name'].lower().replace(' ', '_'),
        'app_name_hyphen': conf['app_name'].lower().replace(' ', '-'),
        'app_version'    : conf['app_version'],
    }
    
    conf['build']['proj_dir'] = path_fmt(
        conf['build']['proj_dir']
    )
    
    conf['build']['dist_dir'] = path_fmt(
        conf['build']['dist_dir'].format(**placeholders)
    )
    
    conf['build']['launcher_name'] = (
        conf['build']['launcher_name'].format(**placeholders)
    )
    
    conf['build']['icon'] = path_fmt(
        conf['build']['icon'] or
        prj_model.python_ico
    )
    
    conf['build']['readme'] = path_fmt(
        conf['build']['readme']
    )
    
    for t in conf['build']['target']:  # type: TTarget
        t['file'] = path_fmt(
            t['file'].format(dist_dir=conf['build']['dist_dir'])
        )
    
    # --------------------------------------------------------------------------
    
    # note: `conf['build']['attachments']` and `conf['build']['module_paths']`
    # may include keyword `dist:xxx`, it requires `path_fmt.dir_o`. so here we
    # must make sure `path_fmt.dir_o` has been defined.
    path_fmt.dir_o = conf['build']['dist_dir']
    
    temp0 = {}  # type: TAttachments
    for k, v in conf['build']['attachments'].items():
        # lk.logt('[D0510]', k, v)
        k = path_fmt(k)
        # noinspection PyUnresolvedReferences
        v = v.split(',')  # type: List[str]
        if v[-1].startswith('dist:'):
            assert len(v) > 1, ((k, v), conf['build']['attachments'])
            path = path_fmt(v[-1].format(name=basename(k)))
            temp0[k] = {'marks': tuple(v[:-1]), 'path': path}
        else:
            temp0[k] = {'marks': tuple(v), 'path': ''}
    conf['build']['attachments'] = temp0
    del temp0
    
    # noinspection PyTypedDict
    conf['build']['attachments_exclusions'] = tuple(
        map(path_fmt, conf['build']['attachments_exclusions'])
    )
    
    # --------------------------------------------------------------------------
    # build:module_paths
    
    module_paths = conf['build']['module_paths']
    for i, p in enumerate(module_paths):
        if p.startswith('dist:'):
            p = path_fmt(p)
        else:
            # FIXME: see `pyportable_installer/main_flow/step3/step3_3/create
            #   _launcher.py > _generate_pylauncher > vars:_ext_paths`
            # p = src_2_dst(
            #     path_fmt(p), conf['build']['proj_dir'], path_fmt.dir_o
            # )
            p = 'src:' + path_fmt(p)
        module_paths[i] = p
    conf['build']['module_paths'] = module_paths
    del module_paths
    
    # # conf['build']['module_paths'] = list(
    # #     map(path_fmt, conf['build']['module_paths'])
    # # )
    
    # --------------------------------------------------------------------------
    # build:venv
    
    mode = conf['build']['venv']['mode']
    options = conf['build']['venv']['options'][mode]
    
    def _load_requirements(req: Union[str, List[str]]):
        if isinstance(req, str):
            if req:
                # FIXME: path_fmt must be 'abspath'
                if path_fmt.fmt == 'relpath':
                    # file = path_fmt.dir_i + '/' + path_fmt
                    return ['...']
                from lk_utils.read_and_write import load_list
                file = path_fmt(req)
                return [x for x in load_list(file)
                        if x and not x.startswith('#')]
            else:
                return []
        else:
            return req
    
    if mode == 'depsland':
        options['venv_name'] = options['venv_name'].format(**placeholders)
        if options['venv_id'] in ('', '_random'):
            options['venv_id'] = str(uuid1()).replace('-', '')
        options['requirements'] = _load_requirements(options['requirements'])
        options['local'] = path_fmt(options['local'])
    elif mode == 'embed_python':
        options['path'] = path_fmt(options['path'])
    elif mode == 'pip':
        options['requirements'] = _load_requirements(options['requirements'])
        options['local'] = path_fmt(options['local'])
    elif mode == 'source_venv':
        options['path'] = path_fmt(options['path'])
    else:
        raise ValueError(mode)
    
    # --------------------------------------------------------------------------
    # build:compiler
    
    name = conf['build']['compiler']['name']
    options = conf['build']['compiler']['options'][name]
    if name == 'pyportable_crypto':
        if options['key'].startswith('path:'):
            assert options['key'].endswith('pyportable_runtime')
            options['key'] = 'path:' + (d := path_fmt(options['key'][5:]))
            assert exists(d)
            lk.logt('[I0946]', '(experimental feature)',
                    'use local precompiled pyportable_runtime package', d)
    
    return conf



from .cython_compiler import CythonCompiler
from .effectless_compiler import EffectlessCompiler
from .mypyc_compiler import MypycCompiler
from .nuitka_compiler import NuitkaCompiler
from .pyarmor_compile import PyArmorCompiler
from .pyc_compiler import PycCompiler
from .pyportable_encryptor import PyportableEncryptor



"""
Step 3.1: copy resources from source to dist.
Step 3.2: compile pyfiles from source and generate compiled (obfuscated)
    results to dist.
Step 3.3: create venv and launcher.
"""
from .build_dist import main



import os
from os import path as xpath
from shutil import copyfile

from lk_logger import lk
from lk_utils import filesniff

from .exclusions import attachments_exclusions_handler
from ....global_conf import gconf
from ....path_model import src_2_dst
from ....typehint import *

__all__ = ['copy_attachments']


def find_files(dir_i):
    for fp, fn in filter(
            lambda x: attachments_exclusions_handler.filter_files(*x),
            filesniff.find_files(dir_i, fmt='zip')
    ):
        yield fp, fn


def find_dirs(dir_i):
    for dp, dn in filter(
            lambda x: attachments_exclusions_handler.filter_dirs(*x),
            filesniff.find_dirs(
                dir_i, fmt='zip', exclude_protected_folders=False
                #                 ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
                #   set this param to False, we will use `exclusions_handler`
                #   to filter the protected folders.
            )
    ):
        yield dp, dn


def _handle_file_exists(file_o, scheme=''):
    """
    Returns:
        str['go_on', 'done']
            'go_on': going on to do the left things.
            'done': all things have done, do not handle this file.
    """
    if scheme == '':  # see `TBuildConf.attachments_exist_scheme`
        scheme = gconf.attachments_exist_scheme
    
    if scheme == 'error':
        raise FileExistsError(file_o)
    elif scheme == 'override':
        os.remove(file_o)
        return 'go_on'
    elif scheme == 'skip':
        return 'done'
    else:
        raise Exception('Unknown scheme', scheme)


_created_dirs = set()


def copy_attachments(attachments: TAttachments) -> Iterator[Tuple[TPath, TPath]]:
    """ Specific for handling attachmets in format of `~.typehint.TAttachments`.
    
    Attachments Marks (`~.typehint.TAttachments._TAttachmentsValue.marks`):
        See complete docs at `~/docs/attachments-marks-demonstration.md`.
        Single Marks:
            asset
            assets
            root_assets
            only_folder
            only_folders
        Double Marks:
            asset,compile
            assets,compile
            root_assets,compile
            
    Args:
        attachments
        
    Notice:
        This function is a generator, if you just want to call and exhaust it,
        use this:
            _ = list(copy_attachments(...))
        See usages in `pyportable_installer.main_flow.step3.step3_3.create_venv
        .create_venv`.
    
    Yields:
        Tuple[src_pyfile, dst_pyfile]
    """
    global _created_dirs
    
    for k, v in attachments.items():
        path_i = k
        path_o = v['path'] or src_2_dst(path_i)
        #   there's no '{name}' placeholder in `path_i` and `path_o` because in
        #   `~.step1.indexing_paths.indexing_paths > attachments related code`
        #   we've handled them.
        marks = v['marks']  # e.g. ('assets', 'compile')
        
        is_yield_pyfile = 'compile' in marks  # type: bool
        #   True: yield pyfile; False: just copy pyfile
        
        # 1. `path_i` is file
        if 'asset' in marks or xpath.isfile(path_i):
            if attachments_exclusions_handler.monitor_transferring(
                    '', path_i, 'file') is False:
                lk.logt('[D4756]', 'the file is in exclusion list', path_i)
            else:
                if (d := xpath.dirname(path_o)) not in _created_dirs:
                    os.makedirs(d, exist_ok=True)
                    _created_dirs.add(d)
                if is_yield_pyfile:
                    yield from _handle_compile(path_i, path_o)
                else:
                    _handle_asset(path_i, path_o)
            continue
        
        # 2. `path_i` is dir
        dir_i = path_i
        dir_o = path_o
        if dir_o not in _created_dirs:
            os.makedirs(dir_o, exist_ok=True)
            _created_dirs.add(dir_o)
        
        if attachments_exclusions_handler.monitor_transferring(
                '', dir_i, 'dir') is False:
            lk.logt('[D4757]', 'the directory is in exclusion list', dir_i)
            continue
        if not xpath.exists(dir_o):
            os.mkdir(dir_o)
        
        if 'root_assets' in marks:
            if is_yield_pyfile:
                yield from _handle_root_assets_and_compile(dir_i, dir_o)
            else:
                _handle_root_assets(dir_i, dir_o)
        
        elif 'assets' in marks:
            if is_yield_pyfile:
                yield from _handle_assets_and_compile(dir_i, dir_o)
            else:
                _handle_assets(dir_i, dir_o)
        
        elif 'only_folders' in marks:
            assert is_yield_pyfile is False
            _handle_only_folders(dir_i, dir_o)
        
        elif 'only_folder' in marks:
            assert is_yield_pyfile is False
            _handle_only_folder(dir_i, dir_o)
        
        else:
            raise ValueError('Unknown or incomplete mark', marks)


# -----------------------------------------------------------------------------

def _handle_assets(dir_i, dir_o):
    _handle_root_assets(dir_i, dir_o)
    for dp, dn in find_dirs(dir_i):
        subdir_i, subdir_o = dp, f'{dir_o}/{dn}'
        os.mkdir(subdir_o)
        # # if not xpath.exists(subdir_o): mkdir(subdir_o)
        _handle_assets(subdir_i, subdir_o)


def _handle_root_assets(dir_i, dir_o):
    for fp, fn in find_files(dir_i):
        file_i, file_o = fp, f'{dir_o}/{fn}'
        copyfile(file_i, file_o)


def _handle_assets_and_compile(dir_i, dir_o):
    yield from _handle_root_assets_and_compile(dir_i, dir_o)
    for dp, dn in find_dirs(dir_i):
        subdir_i, subdir_o = dp, f'{dir_o}/{dn}'
        if not xpath.exists(subdir_o): os.mkdir(subdir_o)
        yield from _handle_assets_and_compile(subdir_i, subdir_o)


def _handle_root_assets_and_compile(dir_i, dir_o):
    for fp, fn in find_files(dir_i):
        file_i, file_o = fp, f'{dir_o}/{fn}'
        if fn.endswith('.py'):  # TODO: ~.endswith(('.py', '.pyw', ...))
            if xpath.exists(file_o) and _handle_file_exists(file_o) == 'done':
                continue
            yield file_i, file_o  # MARK: 20210913113649
        else:
            copyfile(file_i, file_o)


def _handle_only_folders(dir_i, dir_o):
    for dp, dn in find_dirs(dir_i):
        subdir_i, subdir_o = dp, f'{dir_o}/{dn}'
        _handle_only_folder(subdir_i, subdir_o)
        _handle_only_folders(subdir_i, subdir_o)


# noinspection PyUnusedLocal
def _handle_only_folder(dir_i, dir_o):
    if not xpath.exists(dir_o):
        os.mkdir(dir_o)


def _handle_asset(file_i, file_o):
    copyfile(file_i, file_o)


# noinspection PyUnusedLocal
def _handle_compile(file_i, file_o):
    assert file_i.endswith('.py')
    if xpath.exists(file_o) and _handle_file_exists(file_o) == 'done':
        return
    yield file_i, file_o  # MARK: 20210913113657



"""
Placeholders:
    cythonize_packages
    filename
"""
from os import chdir
from os.path import dirname
from sys import path

curr_dir = dirname(__file__)

chdir(curr_dir)
path.append(curr_dir + '/' + r'{cythonize_packages}')

from setuptools import setup  # noqa
from Cython.Build import cythonize  # noqa

setup(ext_modules=cythonize('{filename}', language_level='3'))



import os
import os.path as xpath
from uuid import uuid1

from lk_logger import lk
from lk_utils.filesniff import normpath

from ._env import ASSETS_ENTRY  # str['STANDALONE', 'PACKAGE']

lk.logt('[D2952]', ASSETS_ENTRY)


def src_2_dst(src_path, src_dir='', dst_dir=''):
    if not src_path:
        return ''
    if src_path.startswith('dist:'):
        return src_path.replace('dist:', dst_model.dst_root)
    
    if not src_dir: src_dir = src_model.src_root
    if not dst_dir: dst_dir = dst_model.src_root
    
    p = relpath(src_path, src_dir)
    p = f'{dst_dir}/{p}'
    return p


def relpath(path, start):
    if not path:
        return ''
    return normpath(xpath.relpath(path, start))


class PyPortablePathModel:
    cur_root = normpath(xpath.dirname(__file__))
    
    if ASSETS_ENTRY == 'STANDALONE':
        prj_root = xpath.dirname(cur_root)
    else:
        prj_root = f'{cur_root}/assets'
    
    # prj_root/*
    dist = f'{prj_root}/dist'
    lib = f'{prj_root}/lib'
    temp = f'{prj_root}/temp'
    
    # prj_root/lib/*
    temp_lib = f'{lib}/temp_lib'
    
    # cur_root/*
    checkup = f'{cur_root}/checkup'
    template = f'{cur_root}/template'
    
    # cur_root/template/*
    launch_bat = f'{template}/launch.bat'
    pyarmor = f'{template}/pyarmor'
    pylauncher = f'{template}/pylauncher.txt'
    pyproject = f'{template}/pyproject.json'
    python_ico = f'{template}/python.ico'
    pytransform = f'{template}/pytransform.txt'
    
    # cur_root/template/depsland/*
    _depsland = f'{template}/depsland'
    depsland_launcher_part_a = f'{_depsland}/launcher_part_a.bat'
    depsland_launcher_part_b = f'{_depsland}/launcher_part_b.bat'
    depsland_setup_part_a = f'{_depsland}/setup_part_a.bat'
    depsland_setup_part_b = f'{_depsland}/setup_part_b.txt'
    
    # other
    accessory = f'{cur_root}/compilers/accessory'
    bat_2_exe_converter = f'{cur_root}/bat_2_exe/bat_to_exe_converter.exe'
    cythonize_required_packages_for_python3 = \
        f'{accessory}/cythonize_required_packages_for_python3.zip'
    
    # --------------------------------------------------------------------------
    
    def build_dirs(self):
        if ASSETS_ENTRY == 'STANDALONE':
            assert xpath.exists(self.dist)
            assert xpath.exists(self.lib)
            assert xpath.exists(self.temp_lib)
            assert xpath.exists(self.temp)
        else:
            if not xpath.exists(self.prj_root):
                os.mkdir(self.prj_root)
                os.mkdir(self.dist)
                os.mkdir(self.lib)
                os.mkdir(self.temp_lib)
                os.mkdir(self.temp)
        
        if not xpath.exists(self.temp_lib):
            os.mkdir(self.temp_lib)
        elif os.listdir(self.temp_lib):
            from shutil import rmtree
            rmtree(self.temp_lib)
            os.mkdir(self.temp_lib)
    
    def create_temp_dir(self):
        _temp_dir = f'{self.temp}/{uuid1()}'
        lk.loga('make temporary dir', _temp_dir)
        os.mkdir(_temp_dir)
        return _temp_dir
    
    def get_pyportable_crypto_trial_package(
            self, pyversion, assert_exists=True
    ):
        out = '/'.join((
            self.accessory,
            f'pyportable_crypto_trial_{pyversion}',
            'pyportable_crypto'
        ))
        if assert_exists:
            assert xpath.exists(out), '''
                Currently your requested [python_version][1] is not on the
                [supported trial-list][2].
                Please try the following options to resolve your problem:
                    a) Prompt your requested python_version to "3.8" or "3.9";
                    b) Use a custom pyportable_crypto key instead of trial key;
                       Note: you need to install Microsoft Visual Studio C++
                             Build Tools (2019) on your system.
                    c) Contact pyportable_installer project owner to extend
                       trial keys for requested [python_version][1].
                
                [1]: {0}
                [2]: {1}
            '''.format(
                pyversion,
                f'{self.accessory}/pyportable_crypto_trial_*'
            )
        return out


class SourcePathModel:
    src_root = None
    prj_root = None
    
    # noinspection PyAttributeOutsideInit
    def init(self, src_root, prj_root, **kwargs):
        self.src_root = src_root
        self.prj_root = prj_root
        
        self.readme = kwargs.get('readme', '')
    
    def assert_ready(self):
        assert self.src_root and self.prj_root


class DistributedPathModel:
    """
    Tree:
        dist
        |= hello_world_0.1.0    # dst_root
        |                       # ↑ this node concept doesn't exist in `src_model`
            |= src              # src_root
            |                   # ↑ this is correspondent to `src_model.src_root`
                |= hello_world  # prj_root
                |               # ↑ this is correspondent to `src_model.prj_root`
                    |- main.py
                |- pylauncher.py
    """
    dst_root = None
    src_root = None
    prj_root = None
    
    # noinspection PyAttributeOutsideInit
    def init(self, dst_root: str, prj_relroot: str, **kwargs):
        self.dst_root = dst_root
        self.src_root = f'{self.dst_root}/src'
        self.prj_root = f'{self.src_root}/{prj_relroot}'
        #   prj_relroot: came from
        #       `relpath(src_model.prj_root, src_model.src_root)`
        
        # prj_root/*
        self.pylauncher = f'{self.src_root}/pylauncher.py'
        self.pylauncher_conf = f'{self.src_root}/.pylauncher_conf'
        
        # dst_root/*
        self.build = f'{self.dst_root}/build'
        self.lib = f'{self.dst_root}/lib'
        self.src = f'{self.dst_root}/src'
        self.venv = f'{self.dst_root}/venv'
        
        # dst_root/*/*
        self.manifest = f'{self.build}/manifest.json'
        self.python = f'{self.venv}/python.exe'
        
        # other: depsland
        self.depsland_setup_part_a = f'{self.dst_root}/setup.bat'
        self.depsland_setup_part_b = f'{self.build}/setup.py'
        
        # other: kwargs
        launcher_name = kwargs['launcher_name']
        self.launcher_bat = f'{self.dst_root}/{launcher_name}.bat'
        self.launcher_exe = f'{self.dst_root}/{launcher_name}.exe'
        self.readme = src_2_dst(kwargs.get('readme', ''))
    
    def assert_ready(self):
        assert self.dst_root and self.src_root and self.prj_root


prj_model = PyPortablePathModel()
src_model = SourcePathModel()
dst_model = DistributedPathModel()

prj_model.build_dirs()

__all__ = [
    'src_2_dst', 'relpath',
    'prj_model', 'src_model', 'dst_model'
]



from os.path import abspath
from os.path import dirname

from lk_logger import lk
from lk_utils.filesniff import normpath

from .indexing_paths import indexing_paths
from .init_key_params import init_key_params
from .load_conf import load_conf
from .path_formatter import PathFormatter
from ...typehint import *


def main(pyproj_file: TPath,
         addional_conf: Optional[TConf] = None,
         path_format: TPathFormat = 'abspath',
         _skip_init_key_params=False) -> TConf:
    pyproj_file = normpath(abspath(pyproj_file))
    pyproj_dir = dirname(pyproj_file)
    lk.loga(pyproj_dir)
    
    conf = load_conf(pyproj_file, addional_conf)
    conf = indexing_paths(conf, PathFormatter(pyproj_dir, fmt=path_format))
    if _skip_init_key_params is False:
        init_key_params(conf, pyproj_file=pyproj_file, pyproj_dir=pyproj_dir)
    
    return conf



from typing import Optional

from lk_logger import lk


class AttachmentsExclusions:
    
    def __init__(self):
        self.excluded_dirnames = (
            '__pycache__', '.git', '.idea', '.svn', '.vscode'
        )
        self.excluded_filenames = (
            '.git', '.gitignore', '.gitkeep'
        )
        self.excluded_paths = None  # type: Optional[set[str]]
    
    def _indexing_global_exlusion_list(self):
        from ....global_conf import gconf
        
        if gconf.attachments_exclusions is None:
            lk.logt('[E4050]', '''
                The global attachments exclusions are not initialized!
                You can only import this module after `pyportable_installer
                .main_flow.step1.init_key_params` got executed.
                
                Related:
                    `~.main_flow.step1.init_key_params.init_key_params`
                    `.attachments`
            ''', h='parent')
            raise Exception
        else:
            self.excluded_paths = set(
                p + '/' for p in gconf.attachments_exclusions
            )
        
        lk.logp(
            self.excluded_filenames,
            self.excluded_dirnames,
            self.excluded_paths,
            title='attachments exclusions overview'
        )
    
    def monitor_transferring(self, name: str, path: str, type_: str, h='parent'):
        if self.excluded_paths is None:
            self._indexing_global_exlusion_list()
        assert type_ in ('file', 'dir')
        
        pass_through = True
        
        if (path + '/').startswith(tuple(self.excluded_paths)):
            pass_through = False
        if name:
            if (
                    (type_ == 'file' and name in self.excluded_filenames) or
                    (type_ == 'dir' and name in self.excluded_dirnames)
            ):
                self.excluded_paths.add(path + '/')
                pass_through = False
        
        if not pass_through:
            if type_ == 'file':
                lk.logt('[D5438]', 'skip making file', path, h=h)
            else:
                lk.logt('[D5439]', 'skip making dir', path, h=h)
        return pass_through
    
    def filter_files(self, path: str, name: str):
        return self.monitor_transferring(name, path, 'file', 'grand_parent')
    
    def filter_dirs(self, path: str, name: str):
        return self.monitor_transferring(name, path, 'dir', 'grand_parent')


attachments_exclusions_handler = AttachmentsExclusions()



"""
注意: 本模块在运行时而非编译时生效.
FIXME: 本模块不应放在 `pyportable_installer/compiler` 目录, 请转移.
TODO: this module is under construction.
"""
import importlib
import zipfile
from io import BytesIO

from .base_compiler import BaseRuntimeLoader


class ZipImpLoader(BaseRuntimeLoader):
    """
    References:
        https://www.codenong.com/39135750/
        https://docs.python.org/3/library/zipimport.html
    """
    
    def __init__(self, zip_file):
        with open(zip_file, 'rb') as f:
            # convert in memory bytes to file like object
            data = BytesIO(f.read())
        
        self.zip_file = zip_file
        self.zip = zipfile.ZipFile(data)
        self._paths = tuple(x.filename for x in self.zip.filelist)
    
    def find_module(self, pkg_path: str):
        """ Get path from module.
        
        Args:
            pkg_path: A full package path. e.g. 'aaa.bbb.ccc'
        """
        # Consider `aaa.bbb.ccc` as `aaa/bbb/ccc.py`
        filepath_1 = pkg_path.replace('.', '/') + '.py'
        # Consider `aaa.bbb.ccc` as `aaa/bbb/ccc/__init__.py`
        filepath_2 = pkg_path.replace('.', '/') + '/__init__.py'
        
        if filepath_1 in self._paths:
            return filepath_1, 'module'
        elif filepath_2 in self._paths:
            return filepath_2, 'package'
        else:
            raise ModuleNotFoundError
    
    def load_module(self, pkg_path):
        try:
            file, mark = self.find_module(pkg_path)
        except ModuleNotFoundError:
            raise ImportError(pkg_path)
        
        module = importlib.import_module(file)
        if mark == 'package':
            module.__package__ = pkg_path
        
        return module



from .step3_1 import main as step3_1
from .step3_2 import main as step3_2
from .step3_3 import main as step3_3
from ...typehint import *


def main(conf: TConf):
    """

    Project Structure Example:
        
        # from pyportable_installer.path_model src_model, dst_model
        
        hello_world_project # src_model.src_root
        |= hello_world      # src_model.prj_root
            |- main.py      # source project entry
        |= dist
            |= hello_world_1.0.0    # dst_model.dst_root
                |= build            # dst_model.build
                |= lib              # dst_model.lib
                |= src              # dst_model.src (aka. dst_model.prj_root)
                |= venv             # dst_model.venv
                |- ...
    """
    pyfiles = step3_1(conf)
    
    name = conf['build']['compiler']['name']
    options = conf['build']['compiler']['options'][name]
    step3_2(name, pyfiles, options)
    
    step3_3(conf['build'])


__all__ = ['main']



import os
import re
from os.path import dirname
from os.path import exists

from embed_python_manager import EmbedPythonManager
from lk_logger import lk
from lk_utils import run_cmd_shell
from lk_utils.filesniff import normpath

from ...global_conf import gconf
from ...path_model import prj_model
from ...typehint import *


def init_key_params(conf: TConf, **kwargs):
    # pyproj_file
    gconf.pyproj_file = kwargs['pyproj_file']
    # pyproj_dir
    gconf.pyproj_dir = kwargs['pyproj_dir']
    
    # --------------------------------------------------------------------------
    # attachments
    
    gconf.attachments_exclusions = conf['build']['attachments_exclusions']
    assert type(gconf.attachments_exclusions) is tuple
    gconf.attachments_exist_scheme = conf['build']['attachments_exist_scheme']
    
    # --------------------------------------------------------------------------
    
    pyversion = conf['build']['venv']['python_version']
    assert pyversion, '`conf["build"]["venv"]["python_version"]` not defined'
    pyversion = 'python' + pyversion.replace('.', '')
    gconf.target_pyversion = pyversion
    
    # --------------------------------------------------------------------------
    
    name = conf['build']['compiler']['name']  # type: TCompilerName
    mode = conf['build']['venv']['mode']  # type: TVenvMode
    enable_venv = conf['build']['venv']['enable_venv']  # type: bool
    
    is_full_python_required = False
    if not enable_venv:
        is_full_python_required = True
    elif name in ('cython', 'mypyc', 'nuitka'):
        is_full_python_required = True
    elif name == 'pyportable_crypto':
        key = conf['build']['compiler']['options'][name]['key']
        if key != '_trial' and not key.startswith('path:'):
            is_full_python_required = True
    
    is_embed_python_required = False
    if (
            (name in ('pyarmor', 'pyc')) or
            (enable_venv and mode == 'pip') or
            (enable_venv and mode == 'source_venv' and
             conf['build']['venv']['options']['source_venv']['copy_venv'])
    ):
        is_embed_python_required = True
    
    if name == 'pyportable_crypto':
        key = conf['build']['compiler']['options'][name]['key']
        if key.startswith('path:') and pyversion != 'python39':
            is_embed_python_required = True
    
    if name in ('pyarmor', 'pyc') or \
            (enable_venv and mode in ('pip', 'source_venv')):
        is_embed_python_required = True
    
    is_pip_required = False
    if enable_venv and mode == 'pip':
        is_pip_required = True
    
    lk.logt('[I5239]',
            is_full_python_required,
            is_embed_python_required,
            is_pip_required)
    
    # --------------------------------------------------------------------------
    
    if is_full_python_required:
        # noinspection PyTypedDict
        options = conf['build']['compiler']['options'][name]
        if options['python_path'] in ('_auto_detect', 'auto_detect', ''):
            gconf.full_python = _get_full_python(pyversion)
        else:
            if (p := options['python_path']).endswith('python.exe'):
                lk.logt(
                    '[W2320]',
                    'please do not pass "{}" path to `{}`, use "{}" instead. '
                    'the exe-path support may be removed in the future.'.format(
                        p, 'pyproject.json > build:compiler:options:{}:'
                           'python_path'.format(name), dirname(p)
                    )
                )
                gconf.full_python = p
            else:
                gconf.full_python = p + '/python.exe'
        lk.loga(gconf.full_python)
        assert exists(gconf.full_python)
    
    # --------------------------------------------------------------------------
    
    if is_embed_python_required:
        if is_pip_required:
            gconf.embed_python = _get_embed_python(pyversion, True)
        else:
            gconf.embed_python = _get_embed_python_from_local(pyversion) or \
                                 _get_embed_python(pyversion, False)
        assert exists(gconf.embed_python)


def _get_full_python(pyversion):
    def where_python_installed(pyversion: str):  # pyversion: e.g. 'python39'
        """
        PS: 'full python' means a full featured python, i.e. the regular python
            installed in your computer.
        
        There're two ways to search for full python:
            1. Use `py -{pyversion} ...`
               This is introduced since Python 3.3. See https://docs.python.org
               /3/using/windows.html#launcher for more information.
               Note that there're several inadequacies:
               1. It supports only Python 3.3+.
               2. If user has installed a Python version in user-only position,
                  it doesn't work.
            2. Look for the most frequently used positions. Include system
               environment variables.

        Returns:
            The full python dir. For example: 'C:/Program Files/Python38'
        """
        if pyversion.startswith('python3'):
            try:
                ret = run_cmd_shell(
                    'py -{} -c "import sys;print(sys.executable)"'.format(
                        '3' + '.' + pyversion.replace('python3', '')
                    )
                )
                return normpath(ret).replace('/python.exe', '')
            except:
                pass
        
        for v in filter(None, os.getenv('PATH').lower().split(';')):
            if re.search(rf'[/\\]{pyversion}[/\\]?$', v):
                return normpath(v).rstrip('/')
        
        available_paths = (
            f'{os.path.expanduser("~")}'
            + '/appdata/local/programs/python/{pyversion}',
            f'c:/program files/{pyversion}',
            f'c:/program files (x86)/{pyversion}',
            f'c:/program files (x86)/{pyversion}',
            f'd:/program files/{pyversion}',
            f'd:/program files (x86)/{pyversion}',
            f'd:/program files (x86)/{pyversion}',
            f'e:/program files/{pyversion}',
            f'e:/program files (x86)/{pyversion}',
            f'e:/program files (x86)/{pyversion}',
        )
        
        for p in available_paths:
            if os.path.exists(p):
                return p
        
        raise FileNotFoundError(
            'Auto detection failed finding avaiable executable python in your '
            'computer. Please make sure you have installed the required Python '
            f'(version {pyversion[6:]}) in your computer and fill the path in '
            f'your "~/pyproject.json" file.'
        )
    
    try:
        return where_python_installed(pyversion) + '/python.exe'
    except FileNotFoundError as e:
        lk.logt("[E5951]", 'Cannot find installed python in your computer. '
                           'Please pass in your python path manually (for '
                           'example "C:\\Program Files\\Python38") ...')
        path = input(f'Input path here ({pyversion}): ')
        if not exists(path):
            raise e
        else:
            return normpath(path) + '/python.exe'


def _get_embed_python_from_local(pyversion):
    if pyversion == 'python39' and \
            exists(f'{prj_model.prj_root}/venv/python39._pth.bak'):
        return f'{prj_model.prj_root}/venv/python.exe'
    else:
        return ''


def _get_embed_python(pyversion, add_pip_suits: bool):
    manager = EmbedPythonManager(pyversion)
    manager.change_source('npm_taobao_org.yml')
    manager.deploy(
        add_pip_suits=add_pip_suits,
        add_pip_scripts=False,
        add_tk_suits=False  # FIXME: check tkinter requirement
    )
    return manager.python



import sys

current_pyversion = "python{}{}".format(
    sys.version_info.major, sys.version_info.minor
)
target_pyversion = "python38"
if current_pyversion != target_pyversion:
    raise Exception(
        "Python interpreter version doesn't matched!",
        "Required: {}, got {} ({})".format(
            target_pyversion, current_pyversion, sys.executable
        )
    )

__version__ = "0.2.1-trial"

from .inject import inject
from .inject import encrypt_data




"""
注: 本模块不要有第三方依赖.
"""
from json import load
from os import popen

from .pretty_print import printer


def main(file='manifest.json'):
    """
    Args:
        file: 在发布时, doctor.py 会产生一个编译文件到 {dist}/build/checkup
            .pyc, 在同目录下还有一个 {dist}/build/manifest.json, 就是本参数要传
            入的文件
    
    References:
        template/manifest.json
        prebuild.save_build_info()
    """
    with open(file, 'r', encoding='utf-8') as f:
        conf = load(f)
    
    _check_and_set(conf)
    
    with printer.heading('Successfully finished'):
        input('    Checkup has finished, press enter to close the window. ')


def _check_and_set(conf: dict):
    with printer.heading('Python version checking'):
        target_version = conf['python_version']  # e.g. '3.8'
        current_version, result = check_pyversion(
            *map(int, target_version.split('.'))
            #   target_version: '3.8' -> ['3', '8'] -> (3, 8)
        )
        printer.info(
            f'Target version: >={target_version}',
            f'You version: {current_version}',
            status=result
        )
        if result is False:
            raise EnvironmentError(
                f'Python v{target_version} above is required!'
            )
    
    # --------------------------------------------------------------------------
    
    if conf['check_pip_repo']:
        with printer.heading('Pip repo config checking'):
            info, result = check_pip_conf()
            printer.info(*info.strip().split('\n'), status=result)
            if result is False:
                if printer.ask('The pip configuration is not correct!',
                               'Would you allow the program to fix it?'):
                    _fix_pip_repo_config()
                else:
                    raise EnvironmentError('Pip repo configuration failed')


# ------------------------------------------------------------------------------

def check_pyversion(major, minor, micro=0):
    """ Check Python version.
    
    编译后的 .pyc 文件对 python 版本的要求较高. 例如, 通过 python 3.8.x 编译的
    .pyc 只能通过 python 3.8.x 来运行.
    """
    from sys import version_info as info
    curr_ver = f'{info.major}.{info.minor}.{info.micro}'
    is_fullfilled = False
    if info.major == major:
        if info.minor > minor:
            is_fullfilled = False  # False (default) | True
            #   如果您发布的是源码文件 (.py), 则改为 True
            #   如果您发布的是编译后的文件 (.pyc), 则改为 False
        elif info.minor == minor:
            if info.micro >= micro:
                is_fullfilled = True
    return curr_ver, is_fullfilled


def check_pip_conf():
    """ Check pip configurations.
    
    判断 pip 的仓库地址是否设置为国内镜像.
    """
    ret = popen('pip config list').read()
    return ret, 'https://pypi.tuna.tsinghua.edu.cn/simple' in ret


# ------------------------------------------------------------------------------

def _fix_pip_repo_config():
    with printer.heading('Writing pip config to local pip'):
        printer.info(popen(
            'pip config set global.index-url '
            'https://pypi.tuna.tsinghua.edu.cn/simple'
        ).read())
        printer.info(popen(
            'pip config set global.trusted-host '
            'pypi.tuna.tsinghua.edu.cn'
        ).read())


if __name__ == '__main__':
    main()



from os import makedirs
from os.path import dirname
from os.path import exists
from os.path import isfile
from os.path import relpath

from lk_logger import lk
from lk_utils.filesniff import normpath


class DistTree:
    
    def __init__(self):
        self.paths = []  # type: List[str]
        #   note all elements in `self.paths` are directories. see
        #   implementation in `method:self.add_src_dirs`.
    
    def add_src_dirs(self, *paths: str):
        for p in filter(None, paths):
            assert exists(p), (
                'Source path doesn\'t exist, you may check path spelling '
                'in your pyproject file', p
            )
            if (d := _get_dir(p)) not in self.paths:
                self.paths.append(d)
    
    def suggest_src_root(self):
        try:  # check whether paths are from one hard driver.
            # FIXME: this scheme is only available in Windows platform.
            assert len(set(x.split(':', 1)[0] for x in self.paths)) == 1
        except AssertionError:
            return ''
        else:
            minimal_common_path = min(self.paths, key=lambda p: p.count('/'))
            return dirname(minimal_common_path)
    
    def build_dst_dirs(self, src_root: str, dst_root: str, src_dirs=None):
        """

        Args:
            src_root: see `self.suggest_src_root:returns`. note that the
                `src_root` maybe empty. (the caller doesn't need to do
                anything for it, we will take care of this case.)
            dst_root:
            src_dirs: files which starts with `src_root`.
                None: use `self.paths` instead.
        """
        assert exists(src_root) and exists(dst_root)
        if src_dirs is None:
            src_dirs = self.paths
        
        # part 1.
        existed = {dst_root, }  # set
        
        def _mkdir(dir_):
            if dir_ not in existed:
                existed.add(dir_)
                if not exists(dir_):
                    lk.logt('[D0604]', 'create empty folder', dir_)
                    # mkdir(dir_)
                    makedirs(dir_)
        
        # part 2.
        if src_root == '':
            # FIXME: this is only available in Windows platform.
            src_2_dst = lambda src_path: \
                f'{dst_root}/{src_path.replace(":/", "/")}'
        else:
            src_2_dst = lambda src_path: \
                f'{dst_root}/{normpath(relpath(src_path, src_root))}'
        
        # part 3.
        src_dirs.sort()
        lk.logp(self.paths)
        for src_path in src_dirs:
            dst_path = src_2_dst(src_path)
            _mkdir(dst_path)
        
        del existed
    
    def clear(self):
        self.paths.clear()


def _get_dir(path: str):
    if isfile(path):
        return dirname(path)
    else:
        return path



from os.path import exists
from ..typehint import *


class BaseCompiler:
    
    def __init__(self, python_interpreter):
        """
        Args:
            python_interpreter: str. the path to system installed python dir or
                an embedded python dir. for example 'c:/program files/python39
                /python.exe'.
                by default the embed python is suggested to pass in, which  you
                can use is `pyportable_installer.global_conf.gconf.embed_python`.
                (note that `~.gconf.embed_python` is instantiated after
                `pyportable_installer.main_flow.step1.init_key_params.init_key
                _params`.)
        """
        assert exists(python_interpreter), python_interpreter or '<empty string>'
        self._interpreter = python_interpreter
    
    def compile_all(self, pyfiles: TPyFilesToCompile):
        raise NotImplementedError
    
    def compile_one(self, src_file: TPath, dst_file: TPath):
        raise NotImplementedError


class BaseRuntimeLoader:
    pass



import os.path

from lk_utils.filesniff import normpath


class PathFormatter:
    
    def __init__(self, dir_i, dir_o=None, fmt='abspath'):
        """

        Args:
            dir_i: TPath
            dir_o: Optional[TPath]
                if caller needs to process some paths containing keyword
                'dist:', this param cannot be omitted.
            fmt: Literal['abspath', 'relpath'].
                see abspath usages in `reformat_paths` and relpath usages
                in `aftermath.py:main`.
        """
        self.dir_i = dir_i
        self.dir_o = dir_o
        self.fmt = fmt
    
    def __call__(self, path: str):
        if path == '':
            return ''
        if 'dist:' in path:  # if path.startswith('dist:')
            # FIXME: if self.fmt == 'relpath': ...
            return normpath(path.replace('dist:', self.dir_o + '/'))
        elif os.path.isabs(path):
            if self.fmt == 'abspath':
                return path
            else:
                return normpath(os.path.relpath(path, self.dir_i))
        else:
            if self.fmt == 'abspath':
                return normpath(os.path.abspath(f'{self.dir_i}/{path}'))
            else:
                return path



import shutil

from .base_compiler import BaseCompiler
from ..typehint import TPyFilesToCompile


class EffectlessCompiler(BaseCompiler):
    """ No compiling process, just do copy files.
    
    This is made for debug mode. (See `pyportable_installer.main.debug_build`.)
    """
    
    # noinspection PyMissingConstructor
    def __init__(self):
        del self._interpreter
    
    def compile_all(self, pyfiles: TPyFilesToCompile):
        for i, o in pyfiles:
            shutil.copyfile(i, o)
    
    def compile_one(self, src_file, dst_file):
        shutil.copyfile(src_file, dst_file)



from os import mkdir
from os.path import exists

from lk_logger import lk

from .dist_tree import DistTree
from ...typehint import *


def main(conf: TConf):
    """ Create dist-side tree (all empty folders under `dst_root`) """
    _precheck(conf)
    
    # create main folders under dst_root.
    mkdir(conf['build']['dist_dir'])
    mkdir(conf['build']['dist_dir'] + '/' + 'build')
    mkdir(conf['build']['dist_dir'] + '/' + 'lib')
    mkdir(conf['build']['dist_dir'] + '/' + 'src')
    mkdir(conf['build']['dist_dir'] + '/' + 'src' + '/' + '.pylauncher_conf')
    
    dist_tree = DistTree()
    
    """
    Add to source dirs list:
        conf:
            build:
                + proj_dir
                + target
                + attachments

    Do not add to source dirs list:
        conf:
            build:
                - dist_dir
                - icon
                - readme
                - module_paths
    """
    dist_tree.add_src_dirs(
        conf['build']['proj_dir'],
        *(x['file'] for x in conf['build']['target']),
        *(k for k, v in conf['build']['attachments'].items()
          if v['path'] == ''),
    )
    
    src_root = dist_tree.suggest_src_root()
    dst_root = conf['build']['dist_dir']
    lk.loga(f'the suggest source root directory is: {src_root}')
    
    dist_tree.build_dst_dirs(src_root, f'{dst_root}/src')
    
    # init global path models
    _init_path_models(src_root, dst_root, conf)
    
    return src_root, dst_root


def _precheck(conf: TConf):
    assert not exists(d := conf['build']['dist_dir']), (
        'The target distribution directory ({}) already exists, please appoint '
        'another (non-existent) folder to distribute.'.format(d)
    )
    
    paths_not_exist = []
    for src_path in conf['build']['attachments']:
        if not exists(src_path):
            paths_not_exist.append(src_path)
    if paths_not_exist:
        lk.logp(paths_not_exist)
        raise FileNotFoundError(
            'Please make sure all required paths in `conf["build"]'
            '["attachments"]` are existed.'
        )
    
    # if conf['build']['venv']['enable_venv']:
    #     from .embed_python import EmbedPythonManager
    #     builder = EmbedPythonManager(
    #         pyversion=conf['build']['venv']['python_version']
    #     )
    #     # try to get a valid embed python path, if failed, this method will
    #     # raise an exception to terminate process.
    #     builder.get_embed_python_dir()
    #
    #     mode = conf['build']['venv']['mode']
    #     if mode == 'source_venv':
    #         if venv_path := conf['build']['venv']['options'][mode]['path']:
    #             if venv_path.startswith(src_path := conf['build']['proj_dir']):
    #                 lk.logt('[W2015]', f'''
    #                     Please do not put the Python virtual environment folder
    #                     in your source code folder! This will make the third-
    #                     party libraries to be encrypted, which usually leads to
    #                     unpredicatable errors.
    #                     You can put venv aside with the source code dir, this
    #                     is the recommended parctice.
    #
    #                     Current venv dir: {venv_path}
    #                     Suggest moved to: {ospath.dirname(src_path)}/venv
    #                 ''')
    #                 if input('Continue the process? (y/n): ').lower() != 'y':
    #                     raise SystemExit


def _init_path_models(src_root, dst_root, conf: TConf):
    from ...path_model import dst_model
    from ...path_model import src_model
    from ...path_model import relpath
    
    src_model.init(
        src_root=src_root, prj_root=conf['build']['proj_dir'],
        readme=conf['build']['readme']
    )
    dst_model.init(
        dst_root=dst_root,
        prj_relroot=relpath(src_model.prj_root, src_model.src_root),
        launcher_name=conf['build']['launcher_name'],
        readme=conf['build']['readme']
    )



import os
import shutil
import sys
from textwrap import dedent
from typing import Callable

from lk_logger import lk
from lk_utils import dumps
from lk_utils import loads
from lk_utils import run_cmd_args

from .base_compiler import BaseCompiler
from .cython_compiler import CythonCompiler
from ..global_conf import gconf
from ..path_model import dst_model
from ..path_model import prj_model
from ..path_model import src_model


class PyportableEncryptor(BaseCompiler):
    mode: str  # str['regular', 'random', 'trial', 'delegation']
    _encrypt_data: Callable
    _template: str
    __key: str
    __runtime_dir: str
    
    # noinspection PyMissingConstructor
    def __init__(self, key: str, **_):
        # analyse params
        if key == '_trial':
            mode = 'trial'
        elif key.startswith('path:'):
            mode = 'delegation'
        elif key in ('', '_random'):
            mode = 'random'
        else:
            mode = 'regular'
        
        if mode == 'random':
            from secrets import token_hex
            key = token_hex(32)
        lk.logt('[D2858]', key)
        
        if mode in ('trial', 'delegation'):
            lk.logt('[I2834]',
                    'note: PyportableEncryptor is running under "{}" mode, the '
                    '`key` attribute is invalid.'.format(mode))
        
        # assignments
        self.mode = mode
        self._template = dedent('''
            from pyportable_runtime import inject
            globals().update(inject(globals(), locals(), {ciphertext}))
        ''').strip()  # `pyportable_crypto.inject._validate_source_file`
        self.__key = key
        self.__runtime_dir = key[5:] if mode == 'delegation' else ''
        
        # generate runtime lib
        self._generate_runtime_lib()
        self._encrypt_data = self._load_encryption_func()
    
    def _generate_runtime_lib(self):
        def _gen_init_dot_py():
            # # code = 'from .inject import inject'
            code = dedent('''
                try:
                    import sys
                    
                    current_pyversion = "python{{}}{{}}".format(
                        sys.version_info.major, sys.version_info.minor
                    )
                    
                    target_pyversion = "{0}"
                    
                    if current_pyversion != target_pyversion:
                        raise Exception(
                            "Python interpreter version doesn't matched!",
                            "Required: {{}}, got {{}} ({{}})".format(
                                target_pyversion,
                                current_pyversion,
                                sys.executable
                            )
                        )
                except Exception as e:
                    raise e
                else:
                    del sys
                
                from .inject import inject
            ''').format(gconf.target_pyversion).strip()
            #   note: this code template is copied from `sidework.generate
            #   _pyportable_crypto_trial_version.__generate__init__.<vars:code>`
            dumps(code, f'{dst_dir}/__init__.py')
        
        def _gen_regular_lib():
            try:
                import pyportable_crypto as _crypto
                pyportable_crypto_dir = os.path.dirname(_crypto.__file__)
            except ImportError:
                raise ImportError('Package not found: pyportable_crypto',
                                  '(tip: `pip install pyportable_crypto`)')
            else:
                del _crypto
    
            src_dir = pyportable_crypto_dir
            tmp_dir = prj_model.temp
            
            # 4. generate temporary 'inject.py' in tmp_dir
            code = loads(f'{src_dir}/inject.py')
            code = code.replace('{KEY}', self.__key)  # FIXME
            #   tested on pyportable-crypto v0.1.0, 0.1.1, 0.2.0, 0.2.1
            dumps(code, tmp_file := f'{tmp_dir}/inject.py')
            
            # 5. cythonize from tmp_dir/~.py to dst_dir/~.pyd
            compiler = CythonCompiler(gconf.full_python)
            compiler.compile_one(tmp_file, f'{dst_dir}/inject.pyd')
            
            # 6. cleanup tmp_dir
            compiler.cleanup()
            os.remove(tmp_file)
        
        def _gen_trial_lib():
            shutil.copyfile(
                prj_model.get_pyportable_crypto_trial_package(
                    gconf.target_pyversion) + '/inject.pyd',
                f'{dst_dir}/inject.pyd'
            )
        
        def _gen_prebaked_lib():
            #   this is an experimental feature. see source from
            #   `pyportable_installer.main_flow.step1.indexing_paths
            #   .indexing_paths > the end of lines in the function`
            shutil.copytree(self.__runtime_dir, dst_dir, dirs_exist_ok=True)
        
        # ----------------------------------------------------------------------
        
        # 1. create destination dir
        dst_dir = dst_model.lib + '/pyportable_runtime'
        os.mkdir(dst_dir)
        
        # 2. create '__init__.py' in `pyportable_runtime`
        _gen_init_dot_py()
        
        # 3. generate runtime lib
        if self.mode == 'trial':
            _gen_trial_lib()
        elif self.mode == 'delegation':
            _gen_prebaked_lib()
        else:
            _gen_regular_lib()
    
    def _load_encryption_func(self):
        def _load_regular_encryption():
            import pyportable_crypto
            lk.logt('[D4703]',
                    pyportable_crypto.__version__,
                    pyportable_crypto.__file__)
            from pyportable_crypto import encrypt_data
            return encrypt_data
        
        # noinspection PyUnresolvedReferences
        def _load_prepared_encryption(dir_i):
            """ Load a prepared encryption module from `dir_i`. """
            from secrets import token_hex
            token = 'x' + token_hex(15)
            
            dir_x = f'{prj_model.temp_lib}/{token}'
            dir_o = f'{dir_x}/pyportable_crypto'
            #   notice: `dir_x` cannot be deleted by `pyportable_installer.main
            #   _flow.step4.cleanup._cleanup_intermediate_files`, it will raise
            #   a PermissionError ("Access is denied on '~/temp_lib/~/inject
            #   .pyd' ..."). so we postpone deletion to the next time startup,
            #   let `pyportable_installer.path_model.PyPortablePathModel.build
            #   _dirs` handle it.
            
            os.mkdir(dir_x)
            shutil.copytree(dir_i, dir_o)
            
            # inspired by `lk-lambdex` project
            __PYHOOK__ = {}
            exec(
                dedent('''
                    from sys import path
                    path.append(r'{lib_dir}')
                    from {token} import pyportable_crypto
                    __PYHOOK__['mod'] = pyportable_crypto
                ''').format(
                    lib_dir=prj_model.temp_lib,
                    token=token,
                ),
                {'__PYHOOK__': __PYHOOK__}
            )
            pyportable_crypto = __PYHOOK__['mod']
            # del __PYHOOK__
            
            # lk.logt('[D2009]', dir_i)
            lk.logt('[D4704]',
                    pyportable_crypto.__version__,
                    pyportable_crypto.__file__)
            return pyportable_crypto.encrypt_data
        
        def _load_delegated_encryption():
            lk.logt('[W2908]', '(experimental feature)',
                    'using PyportableRuntimeDelegator to process external '
                    'precompiled pyportable_runtime')
            delegator = PyportableRuntimeDelegator(
                gconf.embed_python, self.__runtime_dir
            )
            setattr(
                self, 'compile_all',
                lambda *args, **kwargs: delegator.compile_all(*args, **kwargs)
            )
            setattr(
                self, 'compile_one',
                lambda *args, **kwargs: delegator.compile_one(*args, **kwargs)
            )
            return None
        
        # ----------------------------------------------------------------------
        
        if self.mode == 'regular':
            return _load_regular_encryption()
        
        elif self.mode == 'trial':
            # TODO: to be explained
            #   note that here we pass `gconf.current_pyversion` instead of
            #   `gconf.target_pyversion` to simplify procedures.
            return _load_prepared_encryption(
                dir_i=prj_model.get_pyportable_crypto_trial_package(
                    gconf.current_pyversion
                )
            )
        
        else:  # self.mode == 'delegation'
            pyversion = loads(f'{self.__runtime_dir}/__pyversion__.txt').strip()
            if pyversion == gconf.current_pyversion:
                return _load_prepared_encryption(dir_i=self.__runtime_dir)
            else:
                return _load_delegated_encryption()
    
    def compile_all(self, pyfiles):
        with lk.counting(len(pyfiles)):
            for i, o in pyfiles:
                lk.logx('compiling', os.path.relpath(i, src_model.prj_root))
                self.compile_one(i, o)
    
    def compile_one(self, src_file, dst_file):
        if os.path.basename(src_file) == '__init__.py':
            shutil.copyfile(src_file, dst_file)
            return dst_file
        
        data = loads(src_file)  # type: str
        data += '\n' + '__PYMOD_HOOK__.update(globals())'
        data = self._encrypt_data(data, self.__key)  # type: bytes
        code = self._template.format(ciphertext=data)  # type: str
        dumps(code, dst_file)
        return dst_file
    
    # @property
    # def key(self):
    #     return self.__key


class PyportableRuntimeDelegator:
    
    def __init__(self, embed_python_exe, pyportable_runtime_dir):
        self._python_exe = embed_python_exe
        self._pyportable_runtime_dir = pyportable_runtime_dir
        self._template = dedent("""
            import sys
            sys.path.insert(0, '{pyportable_runtime_parent_dir}')
            sys.path.append('{lk_utils_dir}')
            
            import os
            import shutil
            from textwrap import dedent
            
            from lk_utils import dumps
            from lk_utils import loads
            from pyportable_runtime import encrypt_data
            
            
            class PyportableEncryptorMini:
            
                def __init__(self):
                    self._encrypt_data = encrypt_data
                    self._template = dedent('''
                        from pyportable_runtime import inject
                        globals().update(inject(globals(), locals(), {{ciphertext}}))
                    ''').strip()
                    # ^ warning: do not modify the template content (e.g. add
                    #   line breaks, adjust whitespaces, etc.), it will violate
                    #   `pyportable_runtime.inject.inject._validate_source_file`
                    self.__key = 'NO_NEED_TO_PASS_A_REAL_KEY'
                
                def compile_all(self, pyfiles):
                    for i, o in pyfiles:
                        self.compile_one(i, o)
                
                def compile_one(self, src_file, dst_file):
                    if os.path.basename(src_file) == '__init__.py':
                        shutil.copyfile(src_file, dst_file)
                        return dst_file
                    
                    data = loads(src_file)  # type: str
                    data += '\\n' + '__PYMOD_HOOK__.update(globals())'
                    data = self._encrypt_data(data, self.__key)  # type: bytes
                    code = self._template.format(ciphertext=data)  # type: str
                    dumps(code, dst_file)
                    return dst_file
            
            
            encryptor = PyportableEncryptorMini()
            encryptor.compile_all({pyfiles})
        """)
    
    def compile_all(self, pyfiles):
        # TODO: use argparse
        code = self._template.format(
            pyportable_runtime_parent_dir=os.path.dirname(
                self._pyportable_runtime_dir
            ),
            lk_utils_dir=os.path.abspath('{}/../lib/site-packages'.format(
                sys.executable
            )),
            pyfiles=pyfiles
        )
        
        dumps(code, x := prj_model.create_temp_dir() +
                         '/delegated_encryption.py')
        run_cmd_args(self._python_exe, x)
    
    def compile_one(self, src_file, dst_file):
        # the `compile_all` is more efficient.
        lk.logt('[W2201]', 'using `PyportableRuntimeDelegator.compile_all` is '
                           'more efficient than `~.compile_one`', h='parent')
        self.compile_all(((src_file, dst_file),))



from shutil import copyfile

from lk_logger import lk

from .attachments import copy_attachments
from ....path_model import dst_model
from ....path_model import src_model
from ....typehint import *


def main(conf: TConf):
    src_model.assert_ready()
    dst_model.assert_ready()
    
    lk.logp(
        (src_model.src_root, src_model.prj_root),
        (dst_model.dst_root, dst_model.src_root, dst_model.prj_root),
        title='model paths overview'
    )
    
    if src_model.readme:
        _create_readme(src_model.readme, dst_model.readme)
    
    pyfiles = []
    pyfiles.extend(_copy_sources())
    pyfiles.extend(copy_attachments(conf['build']['attachments']))
    return pyfiles


def _create_readme(file_i: TPath, file_o: TPath):
    if file_i.endswith('.md'):
        try:
            # TODO: import a markdown_2_html converter.
            #   https://github.com/likianta/markdown_images_to_base64
            from markdown_images_to_base64 import md_2_html_base64
            file_o = file_o.replace('.md', '.html')
            md_2_html_base64(file_i, file_o)
        except ImportError:
            copyfile(file_i, file_o)
    else:
        copyfile(file_i, file_o)


def _copy_sources():
    yield from copy_attachments({
        src_model.prj_root: {'marks': ('assets', 'compile'),
                             'path' : dst_model.prj_root}
    })



from os.path import dirname
from sys import version_info


class GlobalConf:
    """ Most attributes are initialized in `~.main_flow.step1.init_key_params`.
    """
    pyproj_file = ''
    pyproj_dir = ''
    
    current_pyversion = "python{}{}".format(
        version_info.major, version_info.minor
    )
    target_pyversion = ''  # format: 'python**', e.g. 'python38', 'python38-32'
    
    full_python = ''
    embed_python = ''
    
    attachments_exclusions = None
    attachments_exist_scheme = 'error'
    
    @property
    def embed_python_dir(self):
        return dirname(self.embed_python)
    
    @property
    def full_python_dir(self):
        return dirname(self.full_python)


gconf = GlobalConf()

__all__ = ['gconf']



import os
import shutil

from lk_logger import lk
from lk_utils import find_dirs
from lk_utils import find_files
from lk_utils.read_and_write import dumps

from ...path_model import dst_model
from ...path_model import prj_model


def main(pyproj_file, additional_conf):
    _cleanup_intermediate_files()
    _generate_manifest(pyproj_file, dst_model.manifest, additional_conf)


def _cleanup_intermediate_files():
    lk.logt('[I5325]', 'clean up intermediate files')
    
    # wait for bat-2-exe thread joins then delete bat file.
    from ..step3.step3_3.create_launcher import thread_pool
    for bat_file, thread in thread_pool.items():
        thread.join()
        if os.path.exists(bat_file):
            os.remove(bat_file)
    thread_pool.clear()
    
    for d in find_dirs(prj_model.temp):
        shutil.rmtree(d)
        # try:
        #     shutil.rmtree(d)
        # except PermissionError:
        #     continue
    
    for f in find_files(prj_model.temp):
        if f.endswith('.gitkeep'):
            continue
        os.remove(f)


def _generate_manifest(pyproj_file, file_o, additional_conf):
    from ..step1 import main
    conf = main(pyproj_file, additional_conf, 'relpath',
                _skip_init_key_params=True)
    try:
        conf['build']['compiler']['options']['pyportable_crypto']['key'] = ''
        conf['build']['compiler']['options']['zipapp']['password'] = ''
    except KeyError:
        pass
    dumps(conf, file_o)



from typing import *

TPath = str  # abspath
TPathFormat = Literal['abspath', 'relpath']

TPyVersion = str
#   e.g.
#       64bit: '3.6', '3.7', '3.8', etc.
#       32big: '3.6-32', '3.7-32', '3.8-32', etc.
#   see `embed_python/manager.py::EmbedPythonManager.options.keys`

# `template/pyproject.json::build:venv:options:depsland,pip:requirements`
TRequirements = Union[List[str], TPath]

THowVenvCreated = Literal['copy', 'symbolink', 'empty_folder']  # default 'copy'


# class TMisc(TypedDict):
#     """
#     References:
#         `pyportable_installer.main > class:Misc`
#         `pyportable_installer.main_flow.step3.build_dist > func:main`
#     """
#     readme: TPath
#     icon: TPath
#     enable_console: bool
#     copy_checkup_tools: bool
#     create_launch_bat: bool
#     how_venv_created: THowVenvCreated
#     obfuscate_source_code: bool
#     do_aftermath: bool
#     log_level: Literal[0, 1, 2]


class _TDepslandOptions(TypedDict):
    # `template/pyproject.json::build:venv:options:depsland`
    venv_name: str
    venv_id: str
    requirements: TRequirements
    offline: bool
    local: TPath


class _TSourceVenvOptions(TypedDict):
    # `template/pyproject.json::build:venv:options:source_venv`
    path: TPath
    copy_venv: bool  # default True


class _TPipOptions(TypedDict):
    # `template/pyproject.json::build:venv:options:pip`
    requirements: TRequirements
    pypi_url: str
    offline: bool
    local: TPath


class _TEmbedPythonOptions(TypedDict):
    # `template/pyproject.json::build:venv:options:embed_python`
    path: TPath


class _TVenvModeOptions(TypedDict):
    # `template/pyproject.json::build:venv:options`
    # note: keep `this:members:name` sync with `TVenvMode`
    depsland: _TDepslandOptions
    embed_python: _TEmbedPythonOptions
    pip: _TPipOptions
    source_venv: _TSourceVenvOptions
    _no_venv: Optional[dict]  # empty dict or None


# TVenvMode: _TVenvModeOptions:attrs
TVenvMode = Literal['depsland', 'embed_python', 'pip', 'source_venv',
                    '_no_venv']


class TVenvBuildConf(TypedDict):
    # `template/pyproject.json::build:venv`
    enable_venv: bool
    python_version: TPyVersion
    mode: TVenvMode
    options: _TVenvModeOptions


class TTarget(TypedDict):
    # `template/pyproject.json::build:target`
    file: TPath
    function: str
    args: list
    kwargs: dict


# `template/pyproject.json::build:attachments`
# TAttachments = Dict[TPath, str]
_TMark = Literal[
    'asset', 'assets', 'root_assets',
    'compile',
    'only_folder', 'only_folders',
]
_TMarks = Literal[
    ('asset',), ('assets',), ('root_assets',),
    ('only_folder',), ('only_folders',),
    ('asset', 'compile'), ('assets', 'compile'), ('root_assets', 'compile')
]


class _TAttachmentsValue(TypedDict):
    marks: Tuple[_TMark, ...]
    path: TPath


TAttachments = Dict[TPath, _TAttachmentsValue]


class _TCompilerOptions(TypedDict):
    # `template/pyproject.json::build:compiler:compiler_options`
    # note: keep `this:members:name` sync with `TCompilerNames`
    cython: dict
    mypyc: dict
    nuitka: dict
    pyarmor: dict
    pyc: dict
    pyportable_crypto: dict  # {'key': str}
    zipapp: dict
    _no_compiler: Optional[dict]  # empty dict or None


# `template/pyproject.json::build:compiler:name`
TCompilerName = Literal['cython', 'mypyc', 'nuitka', 'pyarmor', 'pyc',
                        'pyportable_crypto', 'zipapp', '_no_compiler']


class TCompiler(TypedDict):
    # `template/pyproject.json::build:compiler`
    name: TCompilerName
    options: _TCompilerOptions


class TExperimentalFeatures(TypedDict):
    add_tkinter: bool
    add_pywin32_support: bool
    alert_dialog_box: Literal['console', 'vbsbox', 'tkinter']
    debug_mode: bool
    reserve_bat_file: bool
    tree_shaking: dict


class TBuildConf(TypedDict):
    # `template/pyproject.json::build`
    proj_dir: TPath
    dist_dir: TPath
    launcher_name: str
    icon: TPath
    target: List[TTarget]
    readme: TPath
    attachments: TAttachments
    attachments_exclusions: Tuple[TPath]
    attachments_exist_scheme: Literal['error', 'override', 'skip']
    module_paths: List[TPath]
    module_paths_scheme: Literal['translate', 'leave as-is']
    venv: TVenvBuildConf
    compiler: TCompiler  # Literal['pyarmor', 'pyc', 'pycrypto']
    experimental_features: TExperimentalFeatures
    enable_console: bool


class TConf(TypedDict):
    # `template/pyproject.json`
    app_name: str
    app_version: str
    description: str
    authors: List[str]
    build: TBuildConf
    note: str
    pyportable_installer_version: str


TPyFilesToCompile = List[Tuple[TPath, TPath]]



from os import path as ospath
from os import remove
from shutil import copyfile
from textwrap import dedent

from lk_logger import lk
from lk_utils import dumps
from lk_utils import loads
from lk_utils import send_cmd
from lk_utils.subproc import new_thread

from .base_compiler import BaseCompiler


class PyArmorCompiler(BaseCompiler):
    
    def __init__(self, python_interpreter):
        super().__init__(python_interpreter)
        
        self._liscense = 'trial'
        #   TODO: see https://pyarmor.readthedocs.io/zh/latest/license.html
        
        # create pyarmor file copy
        origin_file = self._locate_pyarmor_script()
        hacked_file = origin_file.replace('.py', '.py.bak')
        if not ospath.exists(hacked_file):
            content = loads(origin_file)
            dumps(dedent('''\
                import sys
                from os.path import dirname
                sys.path.append(dirname(__file__))
                {content}
            ''').format(content=content), hacked_file)
        self._pyarmor = hacked_file
        self._head = f'"{self._interpreter}" "{self._pyarmor}"'
    
    @staticmethod
    def _locate_pyarmor_script():
        try:
            import pyarmor
        except ImportError as e:
            print('Error: could\'t find pyarmor lib! Please make sure you have '
                  'installed pyarmor (by `pip install pyarmor`).')
            raise e
        file = ospath.abspath(f'{pyarmor.__file__}/../pyarmor.py')
        assert ospath.exists(file)
        return file
    
    def generate_runtime(self, local_dir: str, lib_dir: str):
        from shutil import copytree
        
        dir_i = f'{local_dir}/pytransform'
        dir_o = f'{lib_dir}/pytransform'
        
        if not ospath.exists(dir_i):
            send_cmd(f'{self._head} runtime -O "{local_dir}"')
            #   note the target dir is `local_dir`, not `dir_i`
            #   see `cmd:pyarmor runtime -h`
        copytree(dir_i, dir_o)
    
    def compile_all(self, pyfiles):
        """

        References:
            docs/devnote/how-does-pytransform-work.md
        """
        for src_file, dst_file in pyfiles:
            self.compile_one(src_file, dst_file)
    
    @new_thread
    def compile_one(self, src_file, dst_file):
        """
        Compile `src_file` and generate `dst_file`.

        Args:
            src_file
            dst_file

        References:
            `cmd:pyarmor obfuscate -h`

        Results:
            the `dst_file` has the same content structure:
                from pytransform import pyarmor_runtime
                pyarmor_runtime()
                __pyarmor__(__name__, __file__, b'\\x50\\x59\\x41...')
            `pytransform` comes from `{dist}/lib`, it will be added to
            `sys.path` in the startup (see `pyportable_installer/template/
            bootloader.txt` and `pyportable_installer/no3_build_pyproject.py::
            func:_create_launcher`).

        Notes:
            table of `pyarmor obfuscate --bootstrap {0~4}`

            | command            | result                                      |
            | ================== | =========================================== |
            | `pyarmor obfuscate | each obfuscated file has a header of        |
            |  --bootstrap 0`    | `from .pytransform import pyarmor_runtime`  |
            | ------------------ | ------------------------------------------- |
            | `pyarmor obfuscate | only `__init__.py` has a header of          |
            |  --bootstrap 1`    | `from .pytransform import pyarmor_runtime`  |
            | ------------------ | ------------------------------------------- |
            | `pyarmor obfuscate | each obfuscated file has a header of        |
            |  --bootstrap 2`    | `from pytransform import pyarmor_runtime`   |
            |                    | **this is what we want**                    |
            | ------------------ | ------------------------------------------- |
            | `pyarmor obfuscate | *unknown*                                   |
            |  --bootstrap 3`    |                                             |
            | ------------------ | ------------------------------------------- |
            | `pyarmor obfuscate | *unknown*                                   |
            |  --bootstrap 4`    |                                             |
        """
        lk.loga('compiling', ospath.basename(src_file))
        
        if self._liscense == 'trial':
            # The limitation of content size is 32768 bytes (i.e. 32KB) in
            # pyarmor trial version.
            if (size := ospath.getsize(src_file)) > 32768:
                lk.logt(
                    '[W0357]',
                    f'该文件: "{src_file}" 的体积超出了 pyarmor 试用版的限制 '
                    f'({size} > 32768 Bytes) , 请购买个人版或商业版许可后重新编'
                    f'译! (本文件谨以源码形式打包)'
                )
                copyfile(src_file, dst_file)
                return
        
        cmd = (
            f'{self._head} --silent obfuscate'
            f' -O "{ospath.dirname(dst_file)}"'
            f' --bootstrap 2'
            f' --exact'
            f' --no-runtime'
            f' "{src_file}"'
        )
        #   arguments:
        #       --silent        do not print normal info
        #       --output        output path, pass `dst_file`'s dirname, it will
        #                       generate a compiled file under and has the same
        #                       name with `src_file`
        #       --bootstrap 2   see `docstring:notes:table`
        #       --exact         only obfuscate the listed script(s) (here we
        #                       only obfuscate `src_file`)
        #       --no-runtime    do not generate runtime files (cause we have
        #                       generated runtime files in `{dst}/lib`)
        try:
            send_cmd(cmd)
        except Exception as e:
            lk.logt('[E1747]', 'compile failed', e)
            if ospath.exists(dst_file): remove(dst_file)
            copyfile(src_file, dst_file)



"""
Warnings:
    ~/docs/devnote/currently-known-compilers-issues.md

FIXME:
    This module is not suggest to use, it will be refactored with the guidance
    of `.cython_compiler.CythonCompiler`.
"""
import sys
from os import listdir
from os import mkdir
from os.path import basename
from shutil import copyfile
from shutil import rmtree
from uuid import uuid1

from lk_logger import lk
from lk_utils import find_dirs
from lk_utils import send_cmd
from lk_utils.subproc import run_new_thread

from .base_compiler import BaseCompiler
from ..path_model import prj_model


class NuitkaCompiler(BaseCompiler):
    
    def __init__(self, full_python_interpreter):
        super().__init__(full_python_interpreter)
        self._temp_dir = prj_model.temp
    
    def compile_all(self, pyfiles):
        with lk.counting(len(pyfiles)):
            for i, o in pyfiles:
                o += 'd'  # py -> pyd
                lk.logtx('[D5520]', 'compiling', i, o, h='parent')
                self.compile_one(i, o)
        run_new_thread(self._cleanup)
    
    def compile_one(self, src_file, dst_file):
        """
        Notes:
            Code uses `~/python/scripts/mypyc` source code for reference.
        """
        # assert nuitka package installed
        try:
            # noinspection PyPackageRequirements
            import nuitka
        except ImportError as e:
            raise e
        
        # copy source file from src_dir to tmp_dir
        tmp_dir = f'{self._temp_dir}/{uuid1()}'
        lk.logt('[D1402]', tmp_dir)
        mkdir(tmp_dir)
        tmp_file = f'{tmp_dir}/{basename(src_file)}'
        copyfile(src_file, tmp_file)
        
        # compiling
        send_cmd('{} -m nuitka --module --nofollow-imports --output-dir="{}" '
                 '--no-pyi-file {}'.format(sys.executable, tmp_dir, tmp_file))
        # cmd = subprocess.run([
        #     sys.executable, 'nuitka', '--module', '--nofollow-imports',
        #     '--output-dir="{}"'.format(tmp_dir), '--no-pyi-file', tmp_file
        # ])
        # if cmd.returncode != 0:
        #     lk.logp(tmp_file, cmd, cmd.returncode,
        #             (cmd.stderr or b'').decode(), (cmd.stdout or b'').decode())
        #     raise RuntimeError(tmp_file, cmd)
        
        # get pyd file generated in tmp_dir
        pyd_names = [x for x in listdir(tmp_dir) if x.endswith('.pyd')]
        assert len(pyd_names) == 1
        pyd_name = pyd_names[0]
        pyd_file = f'{tmp_dir}/{pyd_name}'
        
        # copy pyd file from tmp_dir to dst_dir
        copyfile(pyd_file, dst_file)
        
        return dst_file
    
    def _cleanup(self):
        for d in find_dirs(self._temp_dir):
            lk.logt('[D5334]', 'delete dir', d)
            rmtree(d)



"""
Convert jpg, png, etc. to ico.

Requirements:
    pillow
    
BTW: You can find a convertion tool online, for example this site:
    https://www.easyicon.net/covert/
"""


def dialog():
    file_i = input('image: ')
    file_o = file_i.replace('.png', '.ico')
    main(file_i, file_o)


def main(file_i, file_o):
    try:
        # noinspection PyPackageRequirements,PyUnresolvedReferences
        from PIL import Image  # pip install pillow
    except Exception as e:
        print('Please install pillow library (pip install pillow)')
        raise e
    img = Image.open(file_i)
    img.save(file_o)
    img.close()


if __name__ == '__main__':
    dialog()



import os
import os.path as xpath
import pickle
import shutil
from secrets import token_hex

from lk_logger import lk
from lk_utils import dumps
from lk_utils import loads
from lk_utils import ropen
from lk_utils import wopen
from lk_utils.filesniff import get_filename
from lk_utils.subproc import run_new_thread

from .bat_2_exe import bat_2_exe
from ....global_conf import gconf
from ....path_model import *
from ....typehint import Dict
from ....typehint import TBuildConf

thread_pool = {}  # {bat_file: thread, ...}
_is_depsland_mode = False  # TODO: downgrade it to local variable
_abs_paths = ...  # type: dict
_rel_paths = ...  # type: dict


def create_launcher(build: TBuildConf):
    global _is_depsland_mode
    _is_debug_mode = False
    if build['venv']['enable_venv']:
        if build['venv']['mode'] == 'depsland':
            _is_depsland_mode = True
        elif build['venv']['mode'] == '_no_venv':
            _is_debug_mode = True
    
    # depsland setup
    if _is_depsland_mode:
        options = build['venv']['options']['depsland']
        _create_depsland_setup(
            build['launcher_name'],
            options['venv_name'],
            options['venv_id'],
            options['requirements'],
            offline=options['offline'],
            local=options['local'],
            pyversion='python' + build['venv'][
                'python_version'].replace('.', '')
        )
    
    # depsland launcher
    if _is_depsland_mode:
        depsland_launcher_part_a = loads(prj_model.depsland_launcher_part_a)
        depsland_launcher_part_a_temp_dir = prj_model.create_temp_dir()
        # depsland_launcher_part_b = loads(prj_model.depsland_launcher_part_b)
        #   PS: `depsland_launcher_part_b`s content is same with `pyportable
        #   _installer/template/launch.bat`, there's no need to handle it
        #   specifically; just follow the procession of `_create_launcher`.
        depsland_exe_launchers_dir = f'{dst_model.build}/_exe_launchers'
        depsland_bat_launchers_dir = f'{dst_model.build}/_bat_launchers_template'
        if not xpath.exists(depsland_exe_launchers_dir):
            os.mkdir(depsland_exe_launchers_dir)
        if not xpath.exists(depsland_bat_launchers_dir):
            os.mkdir(depsland_bat_launchers_dir)
        if not xpath.exists(f'{dst_model.build}/_bat_launchers'):
            # pyportable_installer/template/depsland/launcher_part_a.bat
            os.mkdir(f'{dst_model.build}/_bat_launchers')
    else:
        depsland_launcher_part_a = None
        depsland_launcher_part_a_temp_dir = None
        depsland_exe_launchers_dir = None
        depsland_bat_launchers_dir = None
    
    # --------------------------------------------------------------------------
    
    deduplicator = LauncherNamesDeduplicator(build['launcher_name'])
    kwargs_part_a = {}
    kwargs_part_b = {
        'add_pywin32_support': build[
            'experimental_features']['add_pywin32_support'],
        'enable_console'     : build['enable_console'],
        'enable_venv'        : build['venv']['enable_venv'],
        'generate_exe'       : False if _is_depsland_mode or
                                        _is_debug_mode else True,
        'module_paths'       : build['module_paths'],
        'module_paths_scheme': build['module_paths_scheme'],
        'venv_python'        : 'python' if _is_debug_mode else '',
    }
    
    for i, t in enumerate(build['target']):
        kwargs_part_a.clear()
        if i == 0:
            kwargs_part_a.update({
                'name'         : build['launcher_name'],
                'icon'         : build['icon'],
                'target'       : t,
                'is_main_entry': True
            })
        else:
            name = get_filename(t['file'], suffix=False)
            name = deduplicator.optimize_name(name, t['function'])
            kwargs_part_a.update({
                'name'         : name,
                'icon'         : '',
                'target'       : t,
                'is_main_entry': False
            })
        
        file = _create_launcher(**kwargs_part_a, **kwargs_part_b)
        
        # ----------------------------------------------------------------------
        
        if _is_depsland_mode:
            # generate exe launcher and put it in `dst_model.build/_exe
            # _launchers`
            file_part_a_i = '{}/{}.bat'.format(
                depsland_launcher_part_a_temp_dir, kwargs_part_a['name']
            )
            file_part_a_o = '{}/{}.exe'.format(
                depsland_exe_launchers_dir, kwargs_part_a['name']
            )
            # update `_abs_paths['bat_file']` and `_abs_paths['exe_file']`,
            # they are global variant and will be fetched by the following
            # `_generate_exe`.
            _abs_paths['bat_file'] = file_part_a_i
            _abs_paths['exe_file'] = file_part_a_o
            dumps(depsland_launcher_part_a.format(
                LAUNCHER_NAME=kwargs_part_a['name']
            ), file_part_a_i)
            _generate_exe(
                icon=kwargs_part_a['icon'],
                enable_console=kwargs_part_b['enable_console']
            )
            
            # move previous bat launcher into `dst_model.build/_bat_launchers`
            file_part_b_i = file
            file_part_b_o = '{}/{}.bat'.format(
                depsland_bat_launchers_dir, kwargs_part_a['name']
            )
            shutil.move(file_part_b_i, file_part_b_o)


def _create_launcher(
        name, icon, target, is_main_entry: bool,
        module_paths=None, module_paths_scheme='translate', **options
):
    """

    Args:
        name:
        icon:
        target:
        is_main_entry:
        **options:

    Keyword Args:
        add_pywin32_support: bool[False]
        enable_console: bool[True]
        enable_venv:
        generate_exe: bool[True]
        module_paths: List[str]
        module_paths_scheme: Literal['translate', 'leave as-is']
        venv_python: str
    """
    launcher_name = name
    conf_name = 'main' if is_main_entry else token_hex(16)
    
    global _abs_paths, _rel_paths
    
    _abs_paths = {
        'target_file': src_2_dst(target['file']),
        'target_dir' : src_2_dst(xpath.dirname(target['file'])),
        'launch_file': dst_model.pylauncher,
        'launch_dir' : dst_model.src_root,
        'conf_dir'   : dst_model.pylauncher_conf,
        'conf_file'  : f'{dst_model.pylauncher_conf}/{conf_name}.pkl',
        'bat_file'   : f'{dst_model.dst_root}/{launcher_name}.bat',
        'exe_file'   : f'{dst_model.dst_root}/{launcher_name}.exe',
    }
    
    _rel0 = lambda p: relpath(p, dst_model.dst_root)  # starts from dst root
    _rel1 = lambda p: relpath(p, dst_model.src_root)  # strats from pylauncher dir
    _rel_paths = {
        'lib_dir'    : _rel1(dst_model.lib),  # '../lib'
        'launch_file': _rel0(_abs_paths['launch_file']),  # 'src/pylauncher.py'
        'target_dir' : _rel1(_abs_paths['target_dir']),  # 'hello_world'
        'conf_file'  : _rel1(_abs_paths['conf_file']),  # '.pylauncher_conf'
        'venv_dir'   : _rel0(dst_model.venv),  # 'venv'
        'venv_python': options.get('venv_python') or
                       _rel0(dst_model.python)  # 'venv/python.exe'
    }
    del _rel0, _rel1
    
    # --------------------------------------------------------------------------
    
    _generate_target_conf(target)
    _generate_pylauncher(module_paths, module_paths_scheme)
    _generate_bat(options.get('enable_venv', True))
    if options.get('generate_exe', True):
        _generate_exe(icon, options.get('enable_console', True))
        return _abs_paths['exe_file']
    else:
        return _abs_paths['bat_file']


def _create_depsland_setup(launcher_name, venv_name, venv_id,
                           requirements, **kwargs):
    shutil.copyfile(
        prj_model.depsland_setup_part_a,
        dst_model.depsland_setup_part_a,
    )
    
    with ropen(prj_model.depsland_setup_part_b) as f:
        template = f.read()
        
        code = template.format(
            LAUNCHER=launcher_name,
            VENV_NAME=venv_name,
            VENV_ID=venv_id,
            PYVERSION=kwargs.get('pyversion', 'python39'),
            REQUIREMENTS=requirements,
            OFFLINE=kwargs.get('offline', False),
            # # LOCAL_DIR=kwargs.get('local', ''),
            LOCAL_DIR=relpath(kwargs.get('local', None), dst_model.build)
        )
    with wopen(dst_model.depsland_setup_part_b) as f:
        f.write(code)


class LauncherNamesDeduplicator:
    """ Make sure there's no duplicate name generated. """
    
    def __init__(self, *top_names):
        self._names_counter = {}  # type: Dict[str, int]
        self._names_counter.update({n: 0 for n in top_names})
    
    def optimize_name(self, name, function=''):
        if name in self._names_counter:
            if function:
                name += ' ({})'.format(function)
                return self.optimize_name(name, '')
            self._names_counter[name] += 1
            name += ' ({})'.format(self._names_counter[name])
        else:
            self._names_counter[name] = 0
        return name


# ------------------------------------------------------------------------------

def _generate_target_conf(target):
    with open(_abs_paths['conf_file'], 'wb') as f:
        target_dir = _rel_paths['target_dir']
        target_pkg = ''
        # target_pkg = target_dir.replace('/', '.')
        target_mod = '{}.{}'.format(
            target_dir.replace('/', '.'),
            get_filename(_abs_paths['target_file'], suffix=False)
        )
        
        pickle.dump({
            'TARGET_DIR'   : target_dir,
            'TARGET_PKG'   : target_pkg,
            'TARGET_MOD'   : target_mod,
            'TARGET_FUNC'  : target['function'],
            'TARGET_ARGS'  : target['args'],
            'TARGET_KWARGS': target['kwargs'],
        }, f)


def _generate_pylauncher(module_paths, module_paths_scheme, **options):
    if module_paths_scheme == 'translate':
        _ext_paths = list(map(
            lambda d: relpath(
                d if not d.startswith('src:') else src_2_dst(d[4:]),
                #   FIXME: this is a temp measure. see source point at
                #       `pyportable_installer/main_flow/step1/indexing_paths.py
                #        > vars:module_paths`
                start=_abs_paths['launch_dir']
            ),
            module_paths
        ))
    else:  # leave module paths as is. (usually this is used in debug mode)
        _ext_paths = list(map(
            lambda d: d[4:] if d.startswith('src:') else d,
            module_paths
        ))
    
    with ropen(prj_model.pylauncher) as f:
        template = f.read()
        code = template.format(
            # see `~/template/pylauncher.txt > docs:placeholders`
            PROJ_LIB_DIR=_rel_paths['lib_dir'],
            ADD_PYWIN32_SUPPORT=str(
                options.get('add_pywin32_support', False)),
            MODULE_PATHS=str(_ext_paths),
            # DEFAULT_CONF='default.pkl',
        )
    
    with wopen(_abs_paths['launch_file']) as f:
        f.write(code)


def _generate_bat(enable_venv):
    with ropen(prj_model.launch_bat) as f:
        template = f.read()
        
        if _is_depsland_mode:
            code = template.format(
                PYTHON='{PYTHON}',  # remain placeholder for the client
                #   side, this value will be filled when user runs depsland
                #   installer.
                #   see `~/template/setup_for_depsland.txt > section
                #   :generate launcher`
                PYLAUNCHER='../../' + _rel_paths['launch_file'],
                #   TODO: to be explained (why we use '../../')
                PYCONF=_rel_paths['conf_file']
            )
            #   '%*' supports passing multiple arguments to python. for
            #   example:
            #       example.bat:
            #           python test.py %*
            #       test.py
            #           import sys
            #           print(sys.argv)
            #       cmd:
            #           example.bat hello world 1 2 3
            #       output in console:
            #           ['test.py', 'hello', 'world', '1', '2', '3']
        elif enable_venv:
            code = template.format(
                PYTHON=_rel_paths['venv_python'].replace('/', '\\'),
                PYLAUNCHER=_rel_paths['launch_file'],
                PYCONF=_rel_paths['conf_file']
            )
        else:
            code = template.format(
                PYTHON=gconf.full_python,
                PYLAUNCHER=_rel_paths['launch_file'],
                PYCONF=_rel_paths['conf_file']
            )
    
    with wopen(_abs_paths['bat_file']) as f:
        f.write(code)


def _generate_exe(icon, enable_console):
    def _run(bat_file, exe_file, icon_file, *options):
        lk.loga('converting bat to exe ... '
                '(it may take several seconds ~ one minute)')
        bat_2_exe(bat_file, exe_file, icon_file, *options)
        lk.loga('convertion bat-to-exe done')
    
    # this is a time-consuming operation (persists 1-10 seconds), we put it
    # in a sub thread_of_bat_2_exe.
    thread_pool[_abs_paths['bat_file']] = run_new_thread(
        _run, _abs_paths['bat_file'], _abs_paths['exe_file'], icon, '/x64',
        '' if enable_console else '/invisible'
    )
    #   the thread_of_bat_2_exe will be recycled in `..step3_4.cleanup`.



# ASSETS_ENTRY:
#   'STANDALONE': distribute self project as a standalone application.
#       see related builder example: `~/examples/dist_pyportable_itself`.
#   'PACKAGE': for poetry build and publish to pypi repo.
#   'AUTO_DETECT': let pyportable-installer detect its environment in runtime.
ASSETS_ENTRY = 'AUTO_DETECT'  # str['AUTO_DETECT', 'STANDALONE', 'PACKAGE']

# ------------------------------------------------------------------------------

if ASSETS_ENTRY == 'AUTO_DETECT':
    from os.path import abspath
    from os.path import basename
    _name = basename(abspath(f'{__file__}/../..'))
    if _name == 'site-packages':
        ASSETS_ENTRY = 'PACKAGE'
    elif _name.startswith(('pyportable_installer', 'pyportable-installer')) \
            or _name == 'src':  # 'src' is appeared when dist as an application.
        ASSETS_ENTRY = 'STANDALONE'
    else:
        raise EnvironmentError('Auto detect failed to locate pyportable-'
                               'installer assets entry point because of '
                               'unknown parent dirname', _name)
assert ASSETS_ENTRY in ('STANDALONE', 'PACKAGE')



from .create_launcher import create_launcher
from .create_venv import create_venv
from ....typehint import TBuildConf


def main(build: TBuildConf):
    if build['venv']['enable_venv']:
        create_venv((m := build['venv']['mode']),
                    build['venv']['options'][m])
    create_launcher(build)



"""
Warnings:
    ~/docs/devnote/currently-known-compilers-issues.md

References:
    https://cython.readthedocs.io/en/latest/src/tutorial/cython_tutorial.html
    https://github.com/TechLearnersInc/cythonizer
    https://www.youtube.com/watch?v=qYgKr_AYNjY

Prerequisites:
    cython package
    windows c compiler (install microsoft visual studio build tools, follow the
        guide of https://www.youtube.com/watch?v=qYgKr_AYNjY)
"""
from os import listdir
from os import mkdir
from os.path import basename
from os.path import exists
from shutil import copyfile
from shutil import rmtree
from textwrap import dedent
from uuid import uuid1

from lk_logger import lk
from lk_utils import dumps
from lk_utils import find_dirs
from lk_utils import run_cmd_args
from lk_utils.subproc import run_new_thread

from .base_compiler import BaseCompiler
from ..path_model import prj_model


class CythonCompiler(BaseCompiler):
    
    def __init__(self, full_python_interpreter):
        """
        Args:
            full_python_interpreter: system installed python (path), for
                example 'c:/program files/python39/python.exe'. usually we pass
                in `~.global_conf.gconf.full_python`.
        """
        super().__init__(full_python_interpreter)
        self._temp_dir = prj_model.temp
        self._template = dedent('''
            # placeholders:
            #   cythonize_packages
            #   filename
            
            from os import chdir
            from os.path import dirname
            from sys import path
            
            curr_dir = dirname(__file__)
            chdir(curr_dir)
            path.append(r'{cythonize_packages}')
            
            from setuptools import setup
            from Cython.Build import cythonize
            
            setup(ext_modules=cythonize('{filename}', language_level='3'))
        ''')
        #   TODO: we may turn to use loading './accessory/setup_for_cythonize
        #       .py' instead of defining long string here in the future.
    
    def compile_all(self, pyfiles):
        with lk.counting(len(pyfiles)):
            for i, o in pyfiles:
                o += 'd'  # *.py -> *.pyd
                lk.logtx('[D5520]', 'compiling', i, o)
                self.compile_one(i, o)
        run_new_thread(self.cleanup)
    
    def compile_one(self, src_file, dst_file):
        filename = basename(src_file)
        
        if filename == '__init__.py':
            # https://stackoverflow.com/questions/58797673/how-to-compile-init
            #   -py-file-using-cython-on-windows
            lk.logt('[D0359]', 'cython doesn\'t compile __init__.py (leave it '
                               'uncompiled, just copy to the dist)', src_file)
            copyfile(src_file, dst_file)
            return dst_file
        
        # copy source file
        tmp_dir = f'{self._temp_dir}/{uuid1()}'
        tmp_file = f'{tmp_dir}/{filename}'
        
        if len(tmp_file) > (path_limit := 150):  # TODO: 150 or lower?
            # the file path is too long, it will cause cythonize failed by
            # 'cl.exe > fatal error code C1081' (see more infomation at
            # https://docs.microsoft.com/en-us/cpp/error-messages/compiler
            # -errors-1/fatal-error-c1081).
            # we have to move `tmp_file` to another place.
            lk.logt('[I3926]', f'the tmp_file path is too long '
                               f'({len(tmp_file)} > {path_limit}), we have to '
                               f'create it in your desktop direcotry (it will '
                               f'be auto removed when compilation done)')
            _desktop_mode = True
            
            # ref: https://blog.csdn.net/u013948858/article/details/75072873
            import os.path
            desktop_dir = os.path.join(os.path.expanduser("~"), 'Desktop')
            assert exists(desktop_dir)
            tmp_dir = f'{desktop_dir}/{uuid1()}'
            tmp_file = f'{tmp_dir}/{filename}'
            assert len(tmp_file) <= path_limit, (
                'The file name is too long to compile!',
                f'{tmp_file = }', f'{src_file = }'
            )
        else:
            _desktop_mode = False
        
        mkdir(tmp_dir)
        copyfile(src_file, tmp_file)
        dumps(self._template.format(
            cythonize_packages=prj_model.cythonize_required_packages_for_python3,
            filename=filename
        ), _setup := f'{tmp_dir}/_setup.py')
        
        from lk_utils.subproc import format_cmd
        lk.logt('[D5200]', format_cmd(
            self._interpreter, _setup, 'build_ext', '--inplace'))
        
        run_cmd_args(self._interpreter, _setup, 'build_ext', '--inplace')
        
        # get pyd file generated in tmp_dir
        pyd_names = [x for x in listdir(tmp_dir) if x.endswith('.pyd')]
        assert len(pyd_names) == 1
        pyd_name = pyd_names[0]
        pyd_file = f'{tmp_dir}/{pyd_name}'
        
        copyfile(pyd_file, dst_file)
        
        if _desktop_mode:
            rmtree(tmp_dir)
        
        return dst_file
    
    def cleanup(self):
        for d in find_dirs(self._temp_dir):
            lk.logt('[D5334]', 'delete dir', d)
            rmtree(d)



from os import path as ospath

from lk_logger import lk

from . import main_flow
from .typehint import TConf
from .typehint import TPath


class Misc:
    """
    DELETE: This class is marked "effectless" in v4.0. For now all its usages
        are restricted within this file. I will totally remove it in v4.1 in
        the future. Some of its features will be merged to pyproject
        configuration file.
    """
    # whether to copy checkup tools (`~/pyportable_installer/checkup`) to dist
    # dir.
    copy_checkup_tools = True
    # whether to create launcher file (exe format).
    create_launch_bat = True
    # how to create venv.
    # note: this option only effects when user enables venv in config file.
    how_venv_created = 'copy'
    # whether to obfuscate source code.
    obfuscate_source_code = True
    # whether to do aftermath work. see `aftermath.py`.
    do_aftermath = True
    # how to print messages when pyportable-installer is processing.
    #   0: disable print function.
    #   1: print main messages (i.e. the built-in `print` behavior).
    #   2: print messages with sourcemap (i.e. default `lk-logger` behavior).
    #   suggestion:
    #       for pyportable-installer self distribution package, set to 0;
    #       for pyportable-installer open source package, set to 1 (default);
    #       for developing and debugging pyportable-installer project, set to 2.
    log_level = 2
    
    @classmethod
    def dump(cls):
        # noinspection PyTypeChecker
        return {
            'copy_checkup_tools'   : cls.copy_checkup_tools,
            'create_launch_bat'    : cls.create_launch_bat,
            'how_venv_created'     : cls.how_venv_created,
            'obfuscate_source_code': cls.obfuscate_source_code,
            'do_aftermath'         : cls.do_aftermath,
            'log_level'            : cls.log_level,
        }


def full_build(file, additional_conf=None):
    Misc.copy_checkup_tools = True
    Misc.create_launch_bat = True
    Misc.how_venv_created = 'copy'
    return main(file, additional_conf or {})


def min_build(file, additional_conf=None):  # FIXME
    Misc.copy_checkup_tools = False
    Misc.create_launch_bat = True
    #   True (suggest) | False
    Misc.how_venv_created = 'empty_folder'
    #   'empty_folder' (suggest) | 'symbolink'
    return main(file, additional_conf or {})


def debug_build(file, additional_conf=None):
    Misc.copy_checkup_tools = False
    Misc.create_launch_bat = True
    Misc.how_venv_created = 'symbolink'
    Misc.obfuscate_source_code = False
    Misc.do_aftermath = False
    Misc.log_verbose = True
    
    if additional_conf is None:
        additional_conf = {'build': {
            'attachments_exist_scheme': 'override',
            'module_paths_scheme': 'leave as-is',
            'compiler': {
                'name': '_no_compiler',
                'options': {'_no_compiler': {}}
            },
            'venv': {
                'mode': '_no_venv',
                'options': {'_no_venv': {}}
            }
        }}
    else:
        node_0 = additional_conf.setdefault('build', {})
        node_0['attachments_exists_scheme'] = 'override'
        node_0['module_paths_scheme'] = 'leave as-is'
        node_a1 = node_0.setdefault('compiler', {})
        node_a1['name'] = '_no_compiler'
        node_a2 = node_a1.setdefault('options', {})
        node_a2['_no_compiler'] = {}
        node_b1 = node_0.setdefault('venv', {})
        node_b1['name'] = '_no_venv'
        node_b2 = node_b1.setdefault('options', {})
        node_b2['_no_venv'] = {}

    return main(file, additional_conf)


def main(pyproj_file: TPath, additional_conf: dict) -> TConf:
    """
    Args:
        pyproj_file: basic configurations in 'pyproject.json'.
        additional_conf: partial congurations to override the basic one's. you
            can pass some dynamic configurations through a build script, this
            is convenient for developers to customize their own dist flow in a
            flexible way, which convers the shortage of using static pyproject
            file.
    
    References:
        '~/docs/devnote/difference-between-roots.md'
    """
    if Misc.log_level == 0:
        lk.disable()
    elif Misc.log_level == 1:
        lk.enable_lite_mode()
    
    conf = main_flow.main(pyproj_file, additional_conf)
    
    m, n = ospath.split(conf['build']['dist_dir'])
    lk.logt("[I2501]", f'See distributed project at: \n\t"{m}:0" >> {n}')
    #   this path link is clickable in pycharm console   ^-----^
    
    lk.enable()
    
    return conf




from lk_utils.read_and_write import loads

from ...typehint import *


def load_conf(pyproj_file: TPath,
              addional_conf: Optional[TConf] = None) -> TConf:
    """

    Args:
        pyproj_file: see template at `~/pyportable_installer/template/pyproject
            .json`
        addional_conf: Optional[dict]. partial TConf.

    References:
        docs/pyproject-template.md
        docs/devnote/difference-between-roots.md > h2:pyproj_root
    """
    conf = loads(pyproj_file)
    if addional_conf:
        _update_additional_conf(conf, addional_conf)
    return conf


def _update_additional_conf(main_conf, additional):
    def _update(node: dict, subject: dict):
        for k, v in node.items():
            if isinstance(v, dict):
                _update(v, subject[k])
            elif isinstance(v, list):
                subject[k].extend(v)
            else:
                subject[k] = v
    
    _update(additional, main_conf)
    return main_conf



"""
Main Flow:
    Step 1: indexing paths in conf.
    Step 2: build paths in dist root.
    Step 3:
        Step 3.1: copy resources from source to dist.
        Step 3.2: compile pyfiles from source and generate compiled (obfuscated)
            results to dist.
        Step 3.3: create venv and launcher.
    Step 4: do aftermath (cleanup intermediate files).
"""
from .step1 import main as step1
from .step2 import main as step2
from .step3 import main as step3
from .step4 import main as step4


def main(pyproj_file, additional_conf):
    conf = step1(pyproj_file, additional_conf)
    step2(conf)
    step3(conf)
    step4(pyproj_file, additional_conf)
    return conf


__all__ = ['main']



from contextlib import contextmanager


class Printer:
    
    def __init__(self, scrwidth=80):
        # scrwidth: 'screen width'. suggest 50|80(default)|120|100.
        self.indent = 0
        self.scrwidth = scrwidth
    
    @contextmanager
    def heading(self, text: str):
        """
        Args:
            text:
            
        Examples:
            Code:
                printer = Printer()
                with printer.heading('This is a heading'):
                    #        ^^^^^^^
                    printer.info('Test OK')
                    printer.info('You need to track the trouble')
            Prints:
                ================ [This is a heading] ================
                    Test OK
                    You need to track the trouble
        """
        text = f'[{text}]'
        assert len(text) <= self.scrwidth, \
            'Please keep the length of text({}) less than {} ' \
            'characters!'.format(len(text) - 2, self.scrwidth - 2)
        self.indent += 4
        delimeter = int((self.scrwidth - len(text)) / 2)
        delimeter = '=' * delimeter
        try:
            print(delimeter, text, delimeter)
            yield self
        finally:
            self.indent -= 4
    
    def info(self, *msg: str, status=None):
        """
        Args:
            status: (None|bool)
                None: No status info
                True: A successful status. It will be translated to '[succeed]'
                False: A failed status. It will be translated to '[failed]'
            *msg:
                Note: Please don't use \n here, it would affect the performance
                of `self._wrap_line`.
                
        Examples:
            Code:
                with printer.heading('HEADER'):
                    printer.info('Alpha', 'Beta', 'Gamma')
            Prints:
                ================ [HEADER] ================
                    Alpha
                    Beta
                    Gamma
            Code:
                with printer.heading('HEADER'):
                    printer.info('Alpha', 'Beta', 'Gamma', status=True)
            Prints:
                ================ [HEADER] ================
                    [succeed]
                        Alpha
                        Beta
                        Gamma
            Code:
                with printer.heading('HEADER'):
                    printer.info('Alpha', 'Beta', 'Gamma', status=False)
            Prints:
                ================ [HEADER] ================
                    [failed]
                        Alpha
                        Beta
                        Gamma
        """
        lspacing = ' ' * self.indent
        
        if isinstance(status, bool):
            print(lspacing + ('[succeed]' if status else '[failed]'))
            lspacing += lspacing
        
        for line in msg:
            print('\n'.join(self._wrap_line(line, lspacing=lspacing)))
        
        return status
    
    def ask(self, *msg):
        """
        
        Args:
            *msg:

        Examples:
            response = printer.ask('The configuration is not correct!',
                                   'Would you like me to repair it?')
        
        Returns:
            resp: bool. Input 'y' or 'Y' returns True, or anything else returns
                False
        """
        lspacing = ' ' * self.indent
        for line in msg[:-1]:
            print('\n'.join(self._wrap_line(line, lspacing=lspacing)))
        return input(lspacing + msg[-1] + ' (y/n): ').lower() == 'y'
    
    def _wrap_line(self, line, lspacing='    '):
        """ Ensure each line not exceeds 80 characters.
                                         ^^ depends on `self.scrwidth`
        
        Notes:
            - 暂不支持根据单词换行. 即换行处可能打破一个完整的单词
        
        Returns:
            (假设 lspacing 是四个空格, 下面为了方便看清, 我们用 · 来表示空格)
            (假设 line 是一个非常长的行 (超过 80 字符), line 的内容将如下所示)
            
            ····Now setup goes to the final step! We will help installing the ne
            ····cessary dependencies into your computer, they will be installed
            ····at the global Python site-packages (usually it can be found at "
            ····C:\\Program Files\\Python38").
        """
        step = self.scrwidth - len(lspacing)
        for i in range(0, len(line), step):
            yield lspacing + line[i: i + step]


printer = Printer()



from textwrap import dedent

from lk_utils import run_cmd_args

from .base_compiler import BaseCompiler


class PycCompiler(BaseCompiler):
    
    def __init__(self, python_interpreter):
        super().__init__(python_interpreter)
        self._template = dedent('''
            from py_compile import compile
            compile('{src_file}', '{dst_file}')
        ''').replace('\n', '\\n').replace("'", "\\'")
    
    def compile_all(self, pyfiles):
        for src_file, dst_file in pyfiles:
            dst_file += 'c'  # *.py -> *.pyc
            self.compile_one(src_file, dst_file)
    
    def compile_one(self, src_file, dst_file):
        """ Compile py file to pyc.

        References:
            https://stackoverflow.com/questions/5607283/how-can-i-manually
                -generate-a-pyc-file-from-a-py-file
            https://docs.python.org/3/library/py_compile.html
        """
        code = self._template.format(src_file=src_file, dst_file=dst_file)
        run_cmd_args(self._interpreter, '-c', f'"exec(\'{code}\')"')
        # run_cmd_shell('"{}" -c "exec(\'{}\')"'.format(self._interpreter, code))



"""
https://github.com/tokyoneon/B2E

Command:
    bat_to_exe_converter /bat {file.bat} /exe {file.exe} {options}

Options:
    see: open 'pyportable_installer/bat_2_exe/bat_to_exe_converter.exe'
         -> menu bar -> tools -> cmd tools

Examples:
    bat_to_exe_converter /bat xxx.bat /exe xxx.exe
    bat_to_exe_converter /bat xxx.bat /exe xxx.exe /icon xxx.ico
    bat_to_exe_converter /bat xxx.bat /exe xxx.exe /x64 /icon xxx.ico
    bat_to_exe_converter /bat xxx.bat /exe xxx.exe /x64 /icon xxx.ico
"""
from os import popen
from os.path import abspath

from lk_utils import send_cmd

_bat_2_exe_converter = abspath(f'{__file__}/../bat_to_exe_converter.exe')


def main(bat_file, exe_file, icon='', *options):
    """
    
    Args:
        bat_file: abspath
        exe_file: abspath
        icon: abspath *.ico
        *options: suggested options below
            /x64
            /invisible
            /password {your_password}
            /uac-admin
            /uac-user
    """
    if icon: assert icon.endswith('.ico')
    cmd = '"{}" /bat "{}" /exe "{}" {} {}'.format(
        _bat_2_exe_converter,
        bat_file,
        exe_file,
        '' if not icon else f'/icon "{icon}"',
        ' '.join(options)
    ).strip()
    return send_cmd(cmd)


if __name__ == '__main__':
    cmd = '{} /bat {} /exe {}'.format(
        _bat_2_exe_converter,
        abspath('../template/launch_by_venv.bat'),
        abspath('../../tests/test1.exe')
    )
    print(cmd)
    print(popen(cmd).read())



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyportable_installer',
 'pyportable_installer.checkup',
 'pyportable_installer.compilers',
 'pyportable_installer.compilers.accessory',
 'pyportable_installer.compilers.accessory.pyportable_crypto_trial_python310.pyportable_crypto',
 'pyportable_installer.compilers.accessory.pyportable_crypto_trial_python38.pyportable_crypto',
 'pyportable_installer.compilers.accessory.pyportable_crypto_trial_python39.pyportable_crypto',
 'pyportable_installer.main_flow',
 'pyportable_installer.main_flow.step1',
 'pyportable_installer.main_flow.step2',
 'pyportable_installer.main_flow.step3',
 'pyportable_installer.main_flow.step3.step3_1',
 'pyportable_installer.main_flow.step3.step3_2',
 'pyportable_installer.main_flow.step3.step3_3',
 'pyportable_installer.main_flow.step3.step3_3.bat_2_exe',
 'pyportable_installer.main_flow.step4']

package_data = \
{'': ['*'], 'pyportable_installer': ['template/*', 'template/depsland/*']}

install_requires = \
['cython',
 'embed-python-manager',
 'fire',
 'lk-logger',
 'lk-utils',
 'pyarmor',
 'pyportable-crypto']

setup_kwargs = {
    'name': 'pyportable-installer',
    'version': '4.1.0',
    'description': 'Build and distribute portable Python application by all-in-one configuration file.',
    'long_description': None,
    'author': 'Likianta',
    'author_email': 'likianta@foxmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""
Warnings:
    ~/docs/devnote/currently-known-compilers-issues.md
    
FIXME:
    This module is not suggest to use, it will be refactored with the guidance
    of `.cython_compiler.CythonCompiler`.
"""
import os
import subprocess
import sys
from os import listdir
from os import mkdir
from os.path import basename
from os.path import dirname
from shutil import copyfile
from shutil import rmtree
from textwrap import dedent
from uuid import uuid1

from lk_logger import lk
from lk_utils import dumps
from lk_utils import find_dirs
from lk_utils.subproc import run_new_thread

from .base_compiler import BaseCompiler
from ..path_model import prj_model


class MypycCompiler(BaseCompiler):
    
    def __init__(self, full_python_interpreter):
        super().__init__(full_python_interpreter)
        self._temp_dir = prj_model.temp
    
    def compile_all(self, pyfiles):
        with lk.counting(len(pyfiles)):
            for i, o in pyfiles:
                o += 'd'  # py -> pyd
                lk.logtx('[D5520]', 'compiling', i, o)
                self.compile_one(i, o)
        run_new_thread(self._cleanup)
    
    def compile_one(self, src_file, dst_file):
        """
        Notes:
            Code uses `~/python/scripts/mypyc` source code for reference.
        """
        # copy source file from src_dir to tmp_dir
        tmp_dir = f'{self._temp_dir}/{uuid1()}'
        lk.logt('[D1402]', tmp_dir)
        mkdir(tmp_dir)
        tmp_file = f'{tmp_dir}/{basename(src_file)}'
        copyfile(src_file, tmp_file)
        
        # create setup.py
        setup_file = f'{tmp_dir}/setup.py'
        setup_code = dedent('''\
            from os import chdir
            from distutils.core import setup
            from mypyc.build import mypycify
            
            chdir(r'{0}')
            
            setup(name='mypyc_output',
                  ext_modules=mypycify({1}, opt_level="{2}"))
        ''').format(tmp_dir, [tmp_file], 3)
        dumps(setup_code, setup_file)
        
        # run setup
        env = os.environ.copy()
        env['PYTHONPATH'] = dirname(sys.executable)
        cmd = subprocess.run(
            [sys.executable, setup_file, 'build_ext', '--inplace'], env=env
        )
        if cmd.returncode != 0:
            raise RuntimeError(cmd, tmp_file)
        
        # get pyd file generated in tmp_dir
        pyd_names = [x for x in listdir(tmp_dir) if x.endswith('.pyd')]
        assert len(pyd_names) == 1
        pyd_name = pyd_names[0]
        pyd_file = f'{tmp_dir}/{pyd_name}'
        
        # copy pyd file from tmp_dir to dst_dir
        copyfile(pyd_file, dst_file)
        
        return dst_file
    
    def _cleanup(self):
        for d in find_dirs(self._temp_dir):
            lk.logt('[D5334]', 'delete dir', d)
            rmtree(d)



import re
from os import makedirs
from os import mkdir

from lk_logger import lk
from lk_utils import dumps
from lk_utils import run_cmd_args

from ..step3_1.attachments import copy_attachments
from ....global_conf import gconf
from ....path_model import dst_model
from ....path_model import prj_model
from ....typehint import TVenvMode


def create_venv(mode: TVenvMode, options):
    if mode == '_no_venv':
        return
    
    elif mode == 'depsland':
        return
    
    elif mode == 'embed_python':
        if options['path']:
            _ = list(copy_attachments({
                options['path']: {
                    'marks': ('assets',),
                    'path' : dst_model.venv
                }
            }))
            # # copytree(options['path'], dst_model.venv)
        else:
            mkdir(dst_model.venv)
    
    elif mode in ('pip', 'source_venv'):
        # note that `copy_attachments` is an iterator, we need to exhaust it to
        # drive it processing.
        _ = list(copy_attachments({
            gconf.embed_python_dir: {
                'marks': ('root_assets',),
                'path' : dst_model.venv
            }
        }))
        
        if mode == 'pip' and options['requirements']:
            python = gconf.embed_python
            requirements = f'{prj_model.temp}/requirements.txt'
            target_dir = dst_model.venv + '/lib/site-packages'
            offline = options['offline'] or ''
            find_links = options['local']
            index_url = options['pypi_url']
            trusted_host = (
                re.search(r'(?:https?://)?([^/]+)', index_url).group(1)
                if index_url else ''
            )
            
            dumps(options['requirements'], requirements)
            makedirs(target_dir)
            run_cmd_args(python, '-m', 'pip', 'install',
                         '-r', requirements,
                         '-t', target_dir,
                         find_links and f'--find-links="{find_links}"',
                         offline and '--no-index',
                         index_url and f'--index-url={index_url}',
                         trusted_host and f'--trusted-host={trusted_host}',
                         '--no-python-version-warning',
                         '--no-input')
        
        elif mode == 'source_venv' and options['path']:
            if options['copy_venv']:
                _ = list(copy_attachments({
                    options['path'] + '/lib/site-packages': {
                        'marks': ('assets',),
                        'path' : dst_model.venv + '/lib/site-packages'
                    }
                }))
                # # copytree(options['path'] + '/lib/site-packages',
                # #          dst_model.venv + '/lib/site-packages')
            else:
                dst_model.venv = options['path']
                #   FIXME: black magic is not recommended!
                lk.logt('[W0644]', 'The source venv is not copied, you may '
                                   'copy it later after progress finished',
                        options['path'])
    
    else:
        raise NotImplemented(mode, options)



from lk_logger import lk

from .... import compilers
from ....global_conf import gconf
from ....typehint import TCompilerName, TPyFilesToCompile


def get_compiler(name: TCompilerName, **kwargs):
    lk.logt('[I1028]', 'compiler name', name)
    if name == 'cython':
        return compilers.CythonCompiler(gconf.full_python)
    elif name == 'mypyc':
        return compilers.MypycCompiler(gconf.full_python)
    elif name == 'nuitka':
        return compilers.NuitkaCompiler(gconf.full_python)
    elif name == 'pyarmor':
        return compilers.PyArmorCompiler(gconf.embed_python)
    elif name == 'pyc':
        return compilers.PycCompiler(gconf.embed_python)
    elif name == 'pyportable_crypto':
        return compilers.PyportableEncryptor(**kwargs)
    elif name == '_no_compiler':
        return compilers.EffectlessCompiler()
    else:
        raise NotImplemented(
            'Unopened compiler in current pyportable version', name
        )


def main(compiler_name: TCompilerName, pyfiles: TPyFilesToCompile, options):
    my_compiler = get_compiler(compiler_name, **options)
    my_compiler.compile_all(pyfiles)
    # del my_compiler
